;; RegReg 1.1, June 2003
;;
;; This is a generator of parsers based on cascade of tagged regular
;; expressions.
;;
;; This code assumes that char->integer and integer->char use ASCII.
;;
;; Copyright (c) 2002, 2003, Mario Latendresse
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;;
;;     * Redistributions of source code must retain the above copyright
;; notice, this list of conditions and the following disclaimer.
;; 
;;     * Redistributions in binary form must reproduce the above
;; copyright notice, this list of conditions and the following disclaimer
;; in the documentation and/or other materials provided with the
;; distribution.
;; 
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;; FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
;; COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;; INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
;; (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
;; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
;; STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
;; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
;; OF THE POSSIBILITY OF SUCH DAMAGE.


;; I: fin, fout, str. File names for input and output.
;;    print-fa?, #t => print the Finite Automaton

(define (regreg-file->scm fin fout print-fa?)
  (let* ((name-dfas      (regreg-file->dfa-vectors fin print-fa?))
	 (name           (car  name-dfas))
	 (dfas           (cadr name-dfas))
	 (vas            `(define ,(string->symbol 
				    (string-append (symbol->string name) "-dfas")) 
			    ,(list 'quasiquote dfas)))
	 (user-before     (caddr  name-dfas))
	 (user-after      (cadddr name-dfas)))

    (with-output-to-file fout
      (lambda ()
	(display-ln ";; File produced by RegReg, version 1.1.")
	(display-ln ";; You should not edit its content. ")
	(newline)
	;; The user defined code before RegReg.
	(display-ln ";; The user defined code before RegReg. ")
	(newline)
	(for-each (lambda (x) (write x) (newline) (newline)) user-before)
	(newline)
	(display-ln ";; The DFAs of all levels.")
	(write vas)
	(newline)
	;; The user defined code after RegReg.
	(display-ln ";; The user defined code after RegReg. ")
	(newline) 
	(for-each (lambda (x) (write x) (newline) (newline)) user-after)
	(newline)
	'completed
	))))

(define (read-all-the-rest)
  (let loop ((o (read)) (r '()))
    (if (not (eof-object? o))
	(loop (read) (cons o r))
	(reverse r))))

;; Read the whole file description and transform the rules in DFA vector form.
;;
;; I: fin, str. The name of the file.
;; O: (name [(errfunc DFA-vector-form)]). The first DFA vector is for the tokens.
;;    name is a symbol, specified in the declare section.
(define (regreg-file->dfa-vectors fin print-fa?)
  (with-input-from-file fin
    (lambda ()
      (let ((before-regreg (read-description)))
	(if (not before-regreg)
	    (err "The file "fin" does not contain an s-expression with head symbol 'RegReg'."))
	(let* ((user-before   (car before-regreg))
	       (description   (cadr before-regreg))
	       (user-after    (read-all-the-rest))
	       (macros        (get-part 'macros description))
	       (declare       (get-part 'declare description))
	       (rule-sections (get-rule-sections description)))
	  
	  (for-each (lambda (e)
		      (if (or (not (pair? e))
			      (not (memv (car e) '(macros level declare))))
			  (err "Do not know what to do with " e)))
		    (cdr description))
	  
	  (if (not (list? declare))
	      (err "No declare level in the RegReg description in file " fin))
	  
	  (if (not (pair? rule-sections))
	      (err "There must be at least one level section in " description))
	  
	  (let* ((name        (get-parm 'name declare)) ;; name must be a symbol.
		 (ms          (macros->macros macros))
		 (nfas        (rule-sections->nfas rule-sections ms))
		 (errfuncs    (map car nfas))
		 (dfas        
		  (mapi (lambda (nfa i) 
			  (display-ln "Starting NFA to DFA for level "(+ i 1)".")
			  (let* ((dfa   (nfa->dfa nfa #t)))
			    (if print-fa? (fa-print dfa))
			    (set-dfa-ops! dfa)
			    dfa))
			(map cadr nfas))))
	    
	    (let ((errs-dfas  (map (lambda (errfunc dfa) 
				     (list 
				      (and errfunc (list 'unquote errfunc))
				      (dfa->vector dfa)))
				   errfuncs dfas)))
	      
	      (for-eachi
	       (lambda (v i) 
		 (display-ln "There are "(- (vector-length (cadr v)) 1)" nodes at level "i"."))
	       errs-dfas
	       1)
	      (list name errs-dfas user-before user-after))
	    ))))))

;; Get to RegReg description. Returns all s-expressions before it and
;; the RegReg description itself.
;;
;; O: (user-before regreg-form) or #f if no regreg description found.
(define (read-description)
  (let loop ((e (read)) (r '()))
    (if (eof-object? e)
	#f
	(cond 
	 ((not (pair? e))         (loop (read) (cons e r)))
	 ((eqv? 'RegReg (car e))  (list (reverse r) e))
	 (else                    (loop (read) (cons e r)))))))
    
;; Return a list of NFAS from the rule-sections. The rule sections are assumed
;; in order from 1 to n.
;;
;; O: [(error-function NFA)]
(define (rule-sections->nfas rule-sections ms)
  (map 
   (lambda (x) (list (car x) (caddr x))) ;; drop the second element
   (mapi-foldl
    (lambda (x r rules-level) 
      (if r
	  (level->nfa x (cadr r) 'token rules-level)
	  ;; The basic tokens are based on characters.
	  (level->nfa x ms 'character rules-level)))
    #f
    rule-sections)))

;; Read the rule sections placing them in the order 1 to n, where n is
;; the number of rule sections. From the input, the rule sections can
;; be in any order.
;;
;; O: [rule-section]
(define (get-rule-sections desc)
  (let loop ((l desc) (r '()))
    (if (pair? l)
	(if (and (pair? (car l)) (eqv? 'level (caar l)))
	    (loop (cdr l) (cons (cdar l) r))
	    (loop (cdr l) r))
	(let* ((n  (length r)) ;; number of rule sections.
	       (v  (make-vector n #f)))
	  (for-each (lambda (rules)
		      (if (or (not (pair? rules))
			      (not (integer? (car rules)))
			      ;; The rule level number must be from 1 to n.
			      (> (car rules) n)
			      ;; The rule level should not be repeated.
			      (vector-ref v (- (car rules) 1)))
			  (err "Wrong level number specified in " rules)
			  (vector-set! v (- (car rules) 1) (cdr rules))))
		    r)
	  (vector->list v)))))

(define (get-part key d)
  (let loop ((l d))
    (if (pair? l)
	(if (and (pair? (car l)) (eqv? key (caar l)))
	    (cdar l)
	    (loop (cdr l)))
	#f)))

;; Transform a level section into an NFA. A level section is a list of
;; lists associating symbols to source-TREs. The source-TREs are
;; transformed into internal form and verified for syntaxical errors.
;; They are transformed into NFAs and all ored together to form one
;; NFA.  An error function may be specified; its name is returned
;; also.
;;
;; I: ts, [(symbol source-TRE)]
;;    ms, macros or list of tokens: (u [(sym sexp-TRE)] [sym])
;;    type, (u character token)
;;    rules-level, int.
;; O: ((u #f (name-error-func list-of-tokens NFA-start-node)
;;     
(define (level->nfa ts ms type rules-level)
  (display-ln "Starting construction of NFA for level "rules-level".")
  (node-init)
  ;; i: the position of the rule
  (let loop ((l ts) (r '()) (i 0) (error-func #f))
    (cond 
     ((not (pair? l)) ;; The end of the level description.
      ;; Create one NFA by oring all rules together.
      (list error-func 
	    (map car (reverse r)) 
	    (nfas->or-nfa (reverse r) rules-level)))
     ((not (pair? (car l)))
      (err "A token description must be a list, but met: " (car l)))
     ((eqv? '<error> (caar l))
      (cond
       ((not (= 2 (length (car l))))
	(err "The <error> clause must have one argument only: " (car l)))
       (error-func (err "The <error> symbol already declared: " error-func))
       ((not (and (pair? (cdar l))  (symbol? (cadar l))))
	(err "The <error> token must be associated with a function name (symbol).")))
      (loop (cdr l) r i (cadar l)))
     ((or (not (symbol? (caar l))) 
	  (not (pair? (cdar l))))
      (err "A token description must be (symbol TRE ...) but met "(car l)" in "ts"."))
     ((assoc (caar l) r) (err "Token " (caar l) " declared twice."))
     (else ;; One rule to transform into a NFA.
      (let* ((e        (car l)) ;; e is a pair.
	     (token-id (car e))
	     (rgexp    (cadr e))
	     (third    (my-list-ref e 2))
	     (fourth   (my-list-ref e 3))
	     (parms    (or (and (pair? third) third) 
			   (and (pair? fourth) fourth)
			   '()))
	     (fun-name (or (and (symbol? third) third) 
			   (and (symbol? fourth) fourth)))
	     (TRE-desc (m-TRE-desc
			i ;; the position of this token in the source.
			token-id
			;; the func to process result.
			(and fun-name (list 'unquote fun-name))
			(if (member 'short parms) 
			    'short 
			    (if (member 'prefer parms) 'prefer 'long))
			(if (member 'discard parms) 'discard #f)
			))
	     (nfa      (sexp-reg->nfa 
			(source-reg->sexp-reg rgexp ms type rules-level token-id)
			TRE-desc))
	     (fs       (pop-final-state nfa))
	     (fn       (m-node 'final TRE-desc))
	     )

	(a-node-e! fs fn 'nop)
	;; Continue to next definition.
	(loop (cdr l) (cons (list (caar l) nfa) r) (+ i 1) error-func))))))

;; This function creates one NFA for all tokens of a rule section.  It
;; performs an or of all NFAs with e-transitions from one new start
;; node. The resulting NFA no longer has the property of only one final node.
;; 
;; I: l, [(token NFA)]. Each NFA has one start and one final nodes.
;; O: start state of one NFA.
(define (nfas->or-nfa token-nfas rules-level)

  ;; O: ([op] [NFA start state]). 
  (define (create-ops-sns token-nfas)
    (let loop ((l token-nfas) (ops '()) (sns '()))
      (if (pair? l)
	  (let* ((token (caar l))
		 (nfa   (cadar l))
		 (op    'nop)
		 (sn    (pop-start-state nfa))
		 )
	    (loop (cdr l) (cons op ops) (cons sn sns)))
	  (list (reverse ops) (reverse sns)))))

  (let* ((rsym (string->symbol 
		(string-append "<level-"(number->string rules-level)">")))
	 (new-start-node (m-node 'start (m-TRE-desc #f rsym #f #f #f)))
	 (ops-sns        (create-ops-sns token-nfas)))
    ;; Create a transition from new-start-node to all NFA start state.
    (for-each
     (lambda (op sn) (a-node-e! new-start-node sn op))
     (car ops-sns) (cadr ops-sns))
    new-start-node))

;; Transform all TRE defined in the macros section into internal form.
;; At the end, no TRE contains any reference to a macro.
;; 
;; I: should be [(symbol TRE)], otherwise it is an error from the user.
;; O: [(sym reg-expr)]
(define (macros->macros macros)
  (let loop ((l macros) (r '()))
    (if (pair? l)
	(begin
	  (if (or (not (list? (car l))) (not (= 2 (length (car l)))))
	      (err "Each macro must be of the form (symbol TRE); found " (car l)))
	  (let* ((id   (caar l))
		 (re   (cadar l))
		 (re2  (source-reg->sexp-reg re r 'character 0 id)))
	    (if (not (rgexp-id? id))
		(err "The macro definition "id" is not a valid id."))
	    (if (not re2)
		(err "The macro regular expression "re" is invalid. "))
	    (if (assv id r)
		(err "The macro "id" is already defined."))
	    (loop (cdr l) (cons (list id re2) r))))
	(reverse r))))

;; I: id, sym.	      
(define (rgexp-id? id)
  (and (symbol? id)
       (for-all (lambda (x) (or (char-alphabetic? x)
				(char-numeric? x) 
				(memv x '(#\_ #\-))))
		(string->list (symbol->string id)))))

;; Transform a TRE from the source file into a standard internal
;; representation.
;;
;; I: rgexp, an TRE from a source file
;;    ms, (u [(str exp)] [sym]). Macro list or symbol list. For the symbol list,
;;        it is the list of tokens recognized by the previous level.
;;    type, (u token character)
;;    rules-level, int.
;;    tk, the token associated with this TRE.
;;
;; O: the start node of a NFA.
;;                 
(define (source-reg->sexp-reg rgexp ms type rules-level tk)

  ;; Internal recursive definition assuming all parameters.
  ;;
  ;; I: internal form TRE.
  ;;    tk, the token-id for the TRE.
  ;; O: a TRE in sexp.
  (define (reg->sexp expr)

    (define (err-reg . strs)
      (err (list-string-append strs)"; in TRE "expr" for token "tk"."))

    (cond
     ((char? expr)   expr)
     ((null? expr)   expr)
     ((string? expr) (string-reg->reg expr ms type rules-level tk))
     ((symbol? expr) 
      ;; it is either a token (level > 1) or a reference to a macro (level 1).
      (case type
	((character) 
	 (if (not (assv expr ms)) (err-reg "Unknown macro: " (symbol->string expr)))
	 (cdr (assv expr ms)))
	((token)
	 (if (not (memv expr ms))
	     (warning "Unknown token '"expr"' from the previous level of tokens: " ms))
	 expr)))
     ((not (list? expr)) (err-reg "Unknown TRE form"))
     (else
      (case (car expr)
	((..)
	 (if (not (= 3 (length expr))) (err-reg ".. needs two operands"))
	 (if (not (and (char? (cadr expr)) (char? (caddr expr))))
	     (err-reg "Operator '..' needs two characters as arguments."))
	 (if (not (char<=? (cadr expr) (caddr expr)))
	     (err-reg "Operator '..': second character must not be smaller than the first.")) 
	 expr)
	((!..)
	 (if (not (= 3 (length expr))) (err-reg "!.. needs two operands"))
	 (if (not (and (char? (cadr expr)) (char? (caddr expr))))
	     (err-reg "Operator '!..' needs two characters as arguments."))
	 (if (not (char<=? (cadr expr) (caddr expr)))
	     (err-reg "Operator '!..': second character must not be smaller than the first.")) 
	 expr)
	((::)
	 (if (p-exists? (lambda (x) (not (char? x))) (cdr expr))
	     (err-reg "Only characters can be specified with operator '::'."))
	 expr)
	((!::) 
	 (if (p-exists? (lambda (x) (not (char? x))) (cdr expr))
	     (err-reg "Only characters can be specified with operator '::'."))
	 expr)
	((?)
	 (if (< (length expr) 2) 
	     (err-reg "Operator '?' needs at least one operand"))
	 (if (= 2 (length expr))
	     (reg->sexp `(: () ,(cadr expr)))
	     (reg->sexp `(: () ,(cdr expr)))))
	;; A binding has the form (= id expr) or (+ id expr1 expr2 ...)
	((=)
	 (if (< (length expr) 3)
	     (err-reg "Operator binding '=' needs at least two operands"))
	 (let* ((id  (cadr expr)))
	   (list '= id (reg->sexp (if (= 3 (length expr))(caddr expr)(cddr expr))))))
	;; positive closure
	((+)
	 (if (< (length expr) 2) (err-reg "Operator '+' needs at least one operand")
	     (list '+ (reg->sexp  (if (= 2 (length expr))
				      (cadr expr)
				      (cdr expr))))))
	;; may contain several elements.
	((*)
	 (if (< (length expr) 2) (err-reg "Operator * needs at least one operand")
	     (list '* (reg->sexp  (if (= 2 (length expr))
				      (cadr expr)
				      (cdr expr))))))
	;; may contain several elements.
	((:)
	 (cond
	  ((< (length expr) 3)(err-reg "Operator ':' needs at least two arguments"))
	  ((= (length expr) 3)
	   (list ': (reg->sexp  (cadr expr)) (reg->sexp  (caddr expr))))
	  (else
	   (reg->sexp `(: ,(cadr expr) (: ,(caddr expr) ,@(cdddr expr)))))))
	(else  ;; assume concatenation
	 (map reg->sexp expr))))))

  (reg->sexp rgexp))

;; Transform a string regular expression into an internal form regular
;; expression. Works for the character based TRE and token based TRE.
;;
;; I: s, str. The source-TRE to translate.
;;    type, (u token character)
;;    ms, (u [(str exp)] [sym]). Macro list or symbol list. For the symbol list,
;;    it is the list of tokens recognized by the previous level.
;;    rules-level, int.
;;
;; O: internal form TRE.
(define (string-reg->reg s ms type rules-level token-id)

  (define (err-reg . strs)
    (err (list-string-append strs)"; in TRE "s" for token '"token-id"'."))

  ;; I: id, str
  (define (search-macro id ms)
    (let* ((macro (assoc id ms)))
      (if macro
	  (cadr macro)
	  (err-reg "The macro '"(symbol->string id)"' is unknown"))))

  (let ((i -1))

    (define (sc) 
      (set! i (+ i 1))
      (if (>= i (string-length s))
	  #f
	  (string-ref s i)))

    (define (sp) 
      (if (>= i (string-length s))
	  #f
	  (if (< i 0)
	      #\space
	      (string-ref s i))))

    ;; back up one character in s.
    (define (unsc)  (set! i (- i 1)))

    (define (end-s?) (= i (string-length s)))

    ;; I: (sp) is '{' or '='.
    ;; O: sym.
    (define (read-id)
      (let loop ((c (sc)) (r '()))
	(cond 
	 ((not c) (err-reg "Premature end of TRE"))
	 ((or (char-alphabetic? c) (char-numeric? c) (memv c '(#\_ #\-)))
	  (loop (sc) (cons c r)))
	 ((memv c '(#\= #\}))
	  (if (null? r) (err-reg "Premature '='  or '}'"))
	  (string->symbol (list->string (reverse r))))
	 (else
	  (err-reg "Invalid char '"(string c)"' in identifier")))))
    
    (define (rgexp-curly plevel clevel)
      (let ((id (read-id)))
	(case (sp)
	  ((#\=)
	   (let ((T (rrgexp->exp '() plevel clevel)))
	     ;; it should be '}'.
	     (if (char=? (sp) #\})
		 `(= ,id ,(mk-concat T))
		 (err-reg "Character '}' expected after macro name '"id"'"))))
	  ((#\})
	   (let ((T (search-macro id ms)))
	     T))
	  (else
	   (err-reg "Character '=' or '}' missing after '{'")))))
    
    (define (rgexp-paren plevel clevel)
      (let ((T (rrgexp->exp '() plevel clevel)))
	(if (or (not (sp)) (not (char=? (sp) #\))))  
	    (err-reg "Character ')' missing after ')'"))
	(if (null? T) (err-reg "Redundant pair of parenthesis"))
	T))

    (define (mk-inset l)
      (if (or (null? l) (null? (cdr l)))
	  l
	  (list (cons ':: l))))

    (define (mk-!inset l)
      (if (null? l)
	  l
	  (list (cons '!:: l))))

    (define (mk-or l)
      (if (list-one-element? l)
	  (car l)
	  `(: ,(car l) ,(mk-or (cdr l)))))

    ;; I: (sp) = '[' 
    (define (rgexp-sqrparen)
      (let loop ((c (sc)) (set '()) (ranges '()) (inv #f))
	(case c
	  ((#f) 
	   (if (memv #\] set)
	       (err-reg "There is a ']' right after '[' without an ending ']'"))
	   (err-reg "Expecting ']' after '['"))
	  ((#\^)
	   (if (and (null? ranges) (null? set))
	       (loop (sc) set ranges #t)
	       (loop (sc) (cons c set) ranges inv)))
	  ((#\])
	   (if (and (null? ranges) (null? set))
	       (loop (sc) (list c) ranges inv)
	       (if inv
		   (car
		    (mk-!inset (delete-duplicates
				(append set 
					(list-append (map range->set ranges))))))
		   (if (null? set)
		       (mk-or ranges)
		       (mk-or `(,@(mk-inset (delete-duplicates set))
				,@ranges))))
	       ))
	  ((#\-)
	   (if (and (null? ranges) (null? set))
	       (loop (sc) (cons #\- set) ranges inv)
	       (let ((c2 (sc)))
		 (loop (sc) (cdr set) (cons `(.. ,(car set) ,c2) ranges)
		       inv))))
	  ((#\\)
	   (let ((c1 (sc)))
	     (if (not c1) 
		 (err-reg "Premature end of square brackets notation '[]'."))
	     (if (char-numeric? c1)
		 (let ((c2 (sc)))
		   (if (and c2 (char-numeric? c2))
		       (let ((c3 (sc)))
			 (if (and c3 (char-numeric? c3))
			     (loop (sc) 
				   (cons (integer->char (string->number (string c1 c2 c3)))
					 set)
				   ranges inv)
			     (loop c3 
				   (cons (integer->char (string->number (string c1 c2)))
					 set)
				   ranges inv
				   )))
		       (loop c2
			     (cons (integer->char (string->number (string c1)))
				   set)
			     ranges inv)))
		 (loop (sc) (cons c1 set) ranges inv))))
	  (else
	   (loop (sc) (cons c set) ranges inv)))))
    
    (define (mk-concat l)
      (if (list-one-element? l)
	  (car l)
	  (reverse l)))

    ;; I: [rgexp]. The REs already parsed. (in reverse order).
    ;;    plevel. Parenthesis level.
    ;; O: [rgexp]. The next parsed RE has been added to the front.
    (define (rrgexp->exp T plevel clevel)
      (case (sc)
	((#f) 
	 (if (> plevel 0) 
	     (err-reg "Missing closing parenthesis ')'"))
	 (if (> clevel 0) 
	     (err-reg "Missing closing brace '}'"))
	 T)
	((#\{) 
	 (let ((T2 (rgexp-curly plevel (+ 1 clevel))))
	   (rrgexp->exp (cons T2 T) plevel clevel)))
	((#\}) (if (= 0 clevel)
		   (err-reg "Missing opening brace '{'")
		   T))
	((#\()
	 (let ((T2 (rgexp-paren (+ 1 plevel) clevel)))
	   (rrgexp->exp (cons (mk-concat T2) T) plevel clevel)))
	((#\)) (if (= plevel 0)
		   (err-reg "Missing opening parenthesis '('")
		   T))
	((#\[)
	 (let ((T2 (rgexp-sqrparen)))
	   (rrgexp->exp (cons T2 T) plevel clevel)))
	((#\.)
	 (rrgexp->exp (cons '(!:: #\newline) T) plevel clevel))
	((#\?)
	 (if (null? T) (err-reg "Operator '?' without TRE on the left"))
	 (rrgexp->exp (cons `(: () ,(car T)) (cdr T)) plevel clevel))
	((#\*)
	 (if (null? T) (err-reg "Operator '*' on empty TRE"))
	 (rrgexp->exp (cons `(* ,(car T)) (cdr T)) plevel clevel))
	((#\+)
	 (if (null? T) (err-reg "Operator '+' on empty TRE"))
	 (rrgexp->exp (cons `(+ ,(car T)) (cdr T)) plevel clevel))
	((#\|)
	 (if (null? T) (err-reg "Premature '|'"))
	 (let ((T2 (rrgexp->exp '() plevel clevel)))
	   (if (null? T2) (err-reg "Premature end of TRE after '|'"))
	   `((: ,(mk-concat T) ,(mk-concat T2)))))
	((#\\)
	 (let ((c1 (sc)))
	   (if (not c1) 
	       (err-reg "The escape char '\\' occurs at the end of TRE"))
	   (if (char-numeric? c1)
	       (let ((c2 (sc)))
		 (if (and c2 (char-numeric? c2))
		     (let ((c3 (sc)))
		       (if (and c3 (char-numeric? c3))
			   (rrgexp->exp (cons (integer->char (string->number (string c1 c2 c3))) T) plevel clevel)
			   (begin
			     (unsc)
			     (rrgexp->exp (cons (integer->char (string->number (string c1 c2))) T) plevel clevel))))
		     (begin
		       (unsc) 
		       (rrgexp->exp (cons (integer->char (string->number (string c1))) T) plevel clevel))))
	       (rrgexp->exp (cons c1 T) plevel clevel))))
	(else ; it is a simple character
	 (rrgexp->exp (cons (sp) T) plevel clevel))))
    
    ;; I: (sp) is '('.
    (define (grammar-rgexp-paren plevel clevel)
      (let ((T (grammar->exp '() plevel clevel)))
	(if (or (not (sp)) (not (char=? (sp) #\))))  
	    (err-reg "Character ')' expected, but found '"(sp)"'"))
	(if (null? T) (err-reg "Redundant pair of parenthesis in grammar TRE"))
	T))

    (define (grammar-rgexp-curly plevel clevel)
      (let ((id (read-id)))
	(case (sp)
	  ((#\=)
	   (let ((T (grammar->exp '() plevel clevel)))
	     ;; it should be '}'.
	     (if (char=? (sp) #\})
		 `(= ,id ,(mk-concat T))
		 (err-reg "Character '}' expected after macro name '"id"'"))))
	  (else
	   (err-reg "Character '=' expected after '{id'")))))

    ;; This function parsed the grammar. It is a regular expression
    ;; where the elements are not characters but tokens.
    ;;
    ;; I: [rgexp]. The REs already parsed. (in reverse order).
    ;; O: [rgexp]. The next parsed RE has been added to the front.
    (define (grammar->exp T plevel clevel)
      (case (sc)
	((#f) T)
	((#\{)
	 (let ((T2 (grammar-rgexp-curly plevel (+ 1 clevel))))
	   (grammar->exp (cons T2 T) plevel clevel)))
	((#\}) (if (= 0 clevel)
		   (err-reg "Missing opening curly brace '{'")
		   T))
	((#\()
	 (let ((T2 (grammar-rgexp-paren (+ plevel 1) clevel)))
	   (grammar->exp (cons (mk-concat T2) T) plevel clevel)))
	((#\)) 
	 (if (= plevel 0)
	     (err-reg "Missing opening parenthesis '('")
	     T))
	((#\?)
	 (if (null? T) (err-reg "Operator '?' without TRE on the left"))
	 (grammar->exp (cons `(: () ,(car T)) (cdr T)) plevel clevel))
	((#\*)
	 (if (null? T) (err-reg "Operator '*' on empty TRE"))
	 (grammar->exp (cons `(* ,(car T)) (cdr T)) plevel clevel))
	((#\+)
	 (if (null? T) (err-reg "Operator '+' on empty TRE"))
	 (grammar->exp (cons `(+ ,(car T)) (cdr T)) plevel clevel))
	((#\|)
	 (if (null? T) (err-reg "Premature '|'"))
	 (let ((T2 (grammar->exp '() plevel clevel)))
	   (if (null? T2) (err-reg "Premature end of TRE after '|'"))
	   `((: ,(mk-concat T) ,(mk-concat T2)))))
	((#\space) (grammar->exp T plevel clevel))
	(else
	 (let ((token (read-token)))
	   (if (not (memv token ms))
	       (err-reg "The token " (symbol->string token)
			  " does not exist in level "
			  (number->string (- rules-level 1))))
	   (unsc)
	   (grammar->exp (cons token T) plevel clevel)))))

    ;; I: (sp) is the beginning of the token, 
    ;;    otherwise it is an error.
    ;; O: str. (sp) points to first character after token.
    (define (read-token)
      (let loop ((c (sp)) (r '()))
	(cond
	 ((or (not c)
	      (not (or (char-alphabetic? c) (char-numeric? c) (memv c '(#\_ #\-)))))
	  (if (null? r)
	      (err-reg "Invalid character for token id, found: " c)
	      (string->symbol (list->string (reverse r)))))
	 (else
	  (loop (sc) (cons c r))))))

    ;; Begin
    (let ((Ts (if (eqv? type 'character) 
		  (rrgexp->exp '() 0 0)
		  (grammar->exp '() 0 0))))
      (mk-concat Ts))))

;; Transform an internal TRE into an NFA. The TRE is assumed well formed.
;;
;; I: rgexp, an s-expr describing a regular expression.
;;    tk, a list describing the TRE options, position, token-id, etc.
;;        see NFA node description.
;; O: the start node of an NFA.
;;                 
(define (sexp-reg->nfa rgexp tk)

  ;; I: internal form TRE.
  ;; O: NFA start node.
  (define (reg->nfa expr)
    (cond
     ((char? expr)
      (let* ((sn   (m-node 'start tk))
	     (fn   (m-node 'final tk)))
	(a-node-tr! `(c ,expr) sn fn 'push)
	sn))
     ((symbol? expr) ;; it is a token of the previous level.
      (let* ((sn   (m-node 'start tk))
	     (fn   (m-node 'final tk)))
	(a-node-tr! `(tk ,expr) sn fn 'push)
	sn))
     ((null? expr)
      ;; epsilon transition. (null string)
      (let* ((sn  (m-node 'start tk))
	     (fn  (m-node 'final tk)))
	(a-node-e! sn fn 'push_empty)
	sn))
     (else
      (case (car expr)
	;; in-range transition
	((..)
	 (let* ((from (cadr expr))
		(to   (caddr expr))
		(sn   (m-node 'start tk))
		(fn   (m-node 'final tk)))
	   (a-node-tr! `(.. ,from ,to) sn fn 'push)
	   sn))
	;; not in-range transition
	((!..)
	 (let* ((from (cadr expr))
		(to   (caddr expr))
		(sn   (m-node 'start tk))
		(fn   (m-node 'final tk)))
	   (a-node-tr! `(!.. ,from ,to) sn fn 'push)
	   sn))
	;; in-set transition
	((::)
	 (let* ((set (cdr expr))
		(sn   (m-node 'start tk))
		(fn   (m-node 'final tk)))
	   (a-node-tr! `(:: ,@set) sn fn 'push)
	   sn))
	;; not in-set transition
	((!::)
	 (let* ((set  (cdr expr))
		(sn   (m-node 'start tk))
		(fn   (m-node 'final tk)))
	   (a-node-tr! `(!:: ,@set) sn fn 'push)
	   sn))
	;; A binding has the form (= id expr) or (+ id expr1 expr2 ...)
	((=)
	 (let* ((id  (cadr expr))
		(x   (reg->nfa (caddr expr)))
		(xf  (pop-final-state x))
		(fn  (m-node 'final tk)))
	   (a-node-e! xf fn `(set ,id))
	   x))
	;; positive closure
	((+)
	 (let* ((x    (reg->nfa  (cadr expr)))
		(sn   (m-node 'start tk))
		(fn   (m-node 'final tk))
		(xss  (pop-start-state x)) 
		(xfs  (pop-final-state x)))
	   (a-node-e! xfs xss 'snoc)       ;; loop back
	   (a-node-e! sn  xss 'push_empty) ;; new start to beginning of x
	   (a-node-e! xfs fn  'snoc)       ;; from final x to new final
	   sn))
	;; may contain several elements.
	((*)
	 (let* ((x    (reg->nfa  (cadr expr)))
		(sn   (m-node 'start tk))
		(fn   (m-node 'final tk))
		(xss  (pop-start-state x)) 
		(xfs  (pop-final-state x)))
	   (a-node-e! sn fn   'push_empty)    ;; from start to final
	   (a-node-e! xfs xss 'snoc)          ;; loop back
	   (a-node-e! sn  xss 'push_empty)    ;;
	   (a-node-e! xfs fn  'snoc)          ;; to final state
	   sn))
	;; may contain several elements.
	((:)
	 (let* ((x1  (reg->nfa  (cadr expr)))
		(x2  (reg->nfa  (caddr expr)))
		(x1s (pop-start-state x1))
		(x2s (pop-start-state x2))
		(x1f (pop-final-state x1))
		(x2f (pop-final-state x2))
		(sn  (m-node 'start tk))
		(fn  (m-node 'final tk)))
	   (a-node-e! sn x1s 'nop)
	   (a-node-e! sn x2s 'nop)
	   (a-node-e! x1f fn 'nop)
	   (a-node-e! x2f fn 'nop)
	   sn))
	(else  ;; assume concatenation
	 (let* ((sn   (m-node 'start tk))
		(fn   (m-node 'final tk)))
	   (let loop ((l expr) (last-node sn))
	     (if (pair? l)
		 (let* ((x   (reg->nfa  (car l)))
			(xs  (pop-start-state x)) 
			(xf  (pop-final-state x)))
		   (if (eq? last-node sn)  ;; first iteration?
		       (a-node-e! last-node xs 'push_empty)
		       (a-node-e! last-node xs 'snoc))
		   (loop (cdr l) xf))
		 (a-node-e! last-node fn 'snoc)))
	   sn))))))

  (reg->nfa rgexp))

;; Transforms a NFA to a DFA. The NFA may have several final states, but
;; only one start state.
;;
;; I: x, NFA node. Should be the start node of NFA.
;;    epsilon-tr, #f => there is no e-tr except on x.
;;    rules-level, int, 1 and up.
;; O: DFA node
(define (nfa->dfa x epsilon-tr)

  ;; Add a-transition from node to an existing DFA node or
  ;; a new DFA node.
  ;; I: node, DFA node. 
  ;;    a, transition code
  ;;    S, R1, R2 [DFA node]. Already created DFA nodes.
  ;; O: #f => no new node
  ;;    otherwise a new DFA node
  (define (new-dfa-node node a)
    (let* ((NFAnodes (if epsilon-tr
			 (e-closure  (node-states-tr a (node-base-nodes node)))
			 (node-states-tr a (node-base-nodes node))))
	   (nodes     (reduce-if-final-and-prefer NFAnodes))
	   (dfa-node  (node-exists? nodes)))
      (if dfa-node
	  (begin
	    (a-node-tr! a node dfa-node 'nop)
	    #f)
	  (let* ((start-no    (nodes->start-node-no nodes))
		 (final-no    (nodes->final-node-no nodes))
		 (new-node    (m-node start-no final-no nodes)))

	    (a-node-tr! a node new-node 'nop)
	    new-node))))

  ;; Create all new transitions trs, out of dfa-node. If the node does not
  ;; exist in (u S R), create it. Return only the new-nodes created.
  ;;
  ;; I: dfa-node, a DFA node
  ;;    trs, [tr-code]. Set of NFA transitions out of dfa-node.
  ;;    S, R, [DFA node], all existing DFA nodes.
  ;; O: [DFA node]
  (define (new-dfa-nodes dfa-node trs)
    (let loop ((l trs) (new-nodes '()))
	(if (pair? l)
	    (let ((new-node (new-dfa-node dfa-node (car l))))
	      (loop (cdr l)
		    (if new-node (cons new-node new-nodes) new-nodes)))
	    new-nodes)))

  ;; Start building DFA
  (node-init)
  (if (not (node-start x))
      (error "nfa->dfa" "x is not a NFA start state" x))

  (let* ((sns        (e-closure  (list x)))
	 (final-node (nodes->final-node-no sns))
	 (dfa-sn     (m-node     (node-no x) final-node sns))
	 (S          (list dfa-sn)))
    ;; S is a set of unvisited nodes.
    (let loop ((S S))
      (if (not (null? S))
	  (let* ((dfa-node  (car S))
		 (trs       (node-states-trs (node-base-nodes dfa-node)))
		 (new-nodes (new-dfa-nodes dfa-node trs)))
	    (loop (append new-nodes (cdr S))))
	  (let () 
	    (display-ln "The number of DFA nodes created: " (node-current-no))
	    dfa-sn
	    )))))

;; ---------------------------------------------------------------
;; This section computes the list of operations (e.g. snoc, push)
;; associated to each DFA arc.
;;

;; Set the operations on the DFA transitions. This is based
;; on the operations of the NFA nodes.
;;
;; I: n, DFA node. This should be the starting node of the DFA.
;; O: n.
(define (set-dfa-ops! n)
  ;; Return all maximal paths of NFA transition operations
  ;; formed by epsilon transitions.
  ;; 
  ;; I: tr-code, a transition code. Not an epsilon transition.
  ;;    nfas, [NFA node]. All NFA nodes forming a DFA node.
  ;; O: [(no no [op])]
  (define (all-paths-for-one-tr tr-code nfas)
    (foldr ;; NFA nodes associated to n
     (lambda (first-nfa-node paths)
       (let* ((trs   (filter (lambda (tr) (tr-in? tr-code (car tr)))
			     (node-trs first-nfa-node))))
	 (if (pair? trs)
	     (let* ((new-paths 
		     (list-append
		      (map (lambda (tr) 
			     (find-e-paths first-nfa-node (cadr tr) (list tr)))
			  trs))))
	       (append new-paths paths))
	     paths)))
     '()
     nfas))

  ;; Go to all DFA nodes, starting at n, without visiting twice the
  ;; same node, setting the operation path for each transition.
  ;;
  ;; I: n, DFA node.
  ;;    S, [DFA node]. Set of visited DFA nodes.
  ;; O: [int]. Set of visited DFA nodes
  (define (set-dfa-ops-rec! n S)
    (let* ((nfas  (node-base-nodes n))
	   (trs   (node-trs n)))
      (foldr ;; transition of DFA n
       (lambda (tr l) ;; tr : (code DFA-node)
	 (let* ((c-DFA  (car tr))
		(to-DFA (cadr tr))
		(paths  (all-paths-for-one-tr c-DFA nfas)))
	   ;; Add operation paths to the transition.
	   (set-car! (cddr tr) paths)
	   (if (not (memq to-DFA l))
	       (set-dfa-ops-rec! to-DFA l)
	       l)))
       (cons n S)
       trs)))

  ;; 
  (display-ln "Starting computation of DFA operation paths.")
  (if (node-start n)
      (let* ((nfa-starting-node (no-nfa-nodes->nfa-node (node-start n) 
							(node-base-nodes n)))
	     (start-ops-paths   (find-e-paths nfa-starting-node
					      nfa-starting-node
					      '())))
	(node-start-set! n start-ops-paths)))

  (set-dfa-ops-rec! n '())
  (display-ln "Ended computation of DFA operation paths.")
  n)

;; Starting from x, find all paths that extends with e-transitions
;; and construct the list of operations.
;;
;; I: first-node, NFA node. The caller consider that this is the first
;;                          node of the path.
;;    x, NFA node.
;;    path, [transition]. Initial path.
;; O: [(no1 no2 [op])]. no1 is the first node number of the path;
;;                      no2 is the last node number of the path.
(define (find-e-paths first-node x path)

  ;; Find the sequence of paths from x having only e-transitions.
  ;;
  ;; I: x, current NFA node.
  ;;    path, [transition]. The current path in construction.
  ;; O: [path]. list of transitions.
  (define (find-rec x path)
    (let* ((etrs  (filter (lambda (tr) (eqv? 'e (car tr)))
			  (node-trs x))))
      (cond
       ((null? etrs) (list path))
       ((and (pair? path) (memq x (map cadr (cdr path))))
	;; a cycle of e-transitions 
	'())
       (else
	(list-append
	 (map (lambda (etr) (find-rec (cadr etr) (cons etr path)))
	      etrs))))))

  ;; I: [op]
  ;; O: [op] where op is not nop.
  (define (delete-nops l)
    (filter (lambda (op) (not (eqv? 'nop op))) l))
  
  (let ((all-paths (find-rec x path)))
    ;; construct [(no-begin no-end [op])]
    (map (lambda (p)
	   (list (node-no first-node) 
		 (node-no (cadar p)) ;; last node is at the beginning
		 (delete-nops (map caddr (reverse p)))))
	 all-paths)))

;; ========NFA======DFA===========================================
;; The DFA and NFA nodes are using the following common structure:
;;
;; node : #(trs start final [NFA-node] no [back-tr] TRE-desc)
;;        trs: [(tr-code DFA-node)]
;;
;; For DFA node:
;;     TRE-desc: #f ???
;;     final : (u NFA-node #f)
;;     start : (u #f [(begin-no end-no [op])]
;;             If it is a start node, it is the list of operation paths.
;;             
;;
;; For NFA state
;;    final: #t or #f => not a final node
;;    TRE-desc : (pos token-id fun-name (u short long prefer) discard?)
;;      discard?: 'discard => do not apply the operations, just
;;                            discard the buffer content.

(define m-node          #t)
(define node-init       #t)
(define node-current-no #t)

(define (node-trs       n)       (vector-ref  n 0))
(define (node-trs-set!  n v)     (vector-set! n 0 v)) 
(define (node-start-set! n v)    (vector-set! n 1 v))
(define (node-start     n)       (vector-ref  n 1))
(define (node-final     n)       (vector-ref  n 2))
(define (node-final-set! n v)    (vector-set! n 2 v))
(define (node-base-nodes n)      (vector-ref  n 3))
(define (node-no        n)       (vector-ref  n 4))
(define (node-back-trs  n)       (vector-ref  n 5))
(define (node-back-trs-set! n v) (vector-set! n 5 v))
(define (node-TRE-desc     n)    (vector-ref  n 6))
(define (node-TRE-desc-set! n v) (vector-set! n 6 v))
(define (node? n) (vector? n))

(define nodes-equal? #f)
(define node-exists? #f)

(define (m-TRE-desc pos token-id fun-name qualifier discard?)
  (list pos token-id fun-name qualifier discard?))
(define (node-final-short? n) (and (node-final n) (node-short? n)))
(define (node-short? n)    (eqv? 'short   (list-ref (node-TRE-desc n) 3)))
(define (node-prefer? n)   (eqv? 'prefer  (list-ref (node-TRE-desc n) 3)))
(define (node-discard? n)  (eqv? 'discard (list-ref (node-TRE-desc n) 4)))
(define (node-tk n)        (cadr (node-TRE-desc n)))
(define (node-pos n)       (car  (node-TRE-desc n)))

(let* ((no  0)  ;; the number of nodes already created.
       (hashNumberNodes 32769)
       (hashv   #f))

  ;; Return an hash value for the nodes.
  (define (nodes->hash nodes)
    ;; (display (cons "nodes->hash " (length nodes)) (current-output-port))
    (let loop ((l nodes) (s 0))
      (if (pair? l)
	  (loop (cdr l) (+ s (node-no (car l))))
	  (modulo s hashNumberNodes))))

  ;; Two nodes are equal if they have the same set of base nodes.
  ;; I: n1, node. nfa-nodes [NFA nodes].
  (set! nodes-equal? 
	(lambda (n1 nfa-nodes)
	  (let ((hashValue1 (nodes->hash (node-base-nodes n1)))
		(hashValue2 (nodes->hash nfa-nodes)))
	    (and (= hashValue1 hashValue2)
		 (set-equal? (node-base-nodes n1) nfa-nodes)))))

  ;; Given a set of base nodes, do they form an existing node?
  (set! node-exists?
	(lambda (nodes)
	  (let* ((hashValue (nodes->hash nodes)))
	    (and (pair? (vector-ref hashv hashValue))
		 (p-exists? (lambda (x) (set-equal? nodes (node-base-nodes x)))
			    (vector-ref hashv hashValue))))))

  (set! node-current-no (lambda () no))
  (set! node-init 
	(lambda () 
	  (set! hashv (make-vector hashNumberNodes '()))
	  (set! no 0)))

  (set! m-node
	;; I: type, int => it is a start node
	;;          #f, it is not a start node,
	;;          (u start final startfinal)
	;;    arg2,  a final node number or #f (DFA), a TRE-desc (NFA);
	;;    nodes, set of nodes representing this node or empty.
	;; O: node
	(lambda (type arg2 . nodes)
	  (set! no (+ 1 no))
	  (let ((node (vector '()        ;; transitions 
			      (if (symbol? type)
				  (if (memv type '(start startfinal)) #t #f)
				  type)
			      (if (integer? arg2)
				  arg2
				  (if (symbol? type)
				      (if (memv type '(final startfinal)) #t #f)
				      arg2))
			      (if (pair? nodes) (car nodes) '())
			      no   ; a unique integer number
			      '()  ; back transitions.
			      ;; NFA node creation provides a TRE description.
			      (and (pair? arg2) arg2)
			      )))
	    ;; Store the node in the hash table.
	    (let* ((hashValue (or (and (pair? nodes) (nodes->hash (car nodes)))
				  (modulo no hashNumberNodes))))
	      ;; (if (pair? (vector-ref hashv hashValue))
	      ;; (display (list "Hash hit! at " hashValue " " (length (vector-ref hashv hashValue))) (current-error-port)))
	      (vector-set! hashv hashValue 
			   (cons node  (vector-ref hashv hashValue))))
	    node
	    ))))

;; Add a transition from node n1 to n2.
;; I: a, a transition code.
;;    n1, n2, nodes.
;; O: n1
(define (a-node-tr! a n1 n2 op)
  (node-trs-set! n1 (cons (list a n2 op) (node-trs n1)))
  (node-back-trs-set! n2 (cons (list a n1 #f) (node-back-trs n2)))
  n1)

;; I: n, node.
;; O: [node]. Nodes visited from n.
(define (fa-print n)
  ;; I: S, [node]. Already printed nodes.
  ;; O: [node]. The printed nodes, including S.
  (define (fa-print-rec n S)
    (if (not (memq n S))
	(let* ((transitions 
		(map (lambda (tr) (list (car tr) 
					(node-no (cadr tr)) 
					(caddr tr)))
		     (node-trs n))))
	  (display-l "Node no: " (node-no n) "  ")
	  (write-l transitions)
	  (if (node-final n) 
	      (if (node? (node-final n))
		  (display-l " final " (node-no (node-final n)) ", ")
		  (display-l " final " (node-final n)", ")))
	  (if (node-start n) (display-l " start, " (node-start n)))
	  (if (pair? (node-base-nodes n))
	      (begin (display " base-nodes : ")
		     (for-each (lambda (x) (display-l (node-no x) " "))
			       (node-base-nodes n))))
	  (newline)
	  (foldr (lambda (x l) (fa-print-rec x l))
		 (cons n S)
		 (map cadr (node-trs n))))
	S))
  (fa-print-rec n '()))

;; I: l1, l2, [node]
(define (nodes-union l1 l2)
  ;; (display "nodes-union " (current-error-port))
  (let loop ((l (append l1 l2)) (r '()))
    (if (pair? l)
	(if (memq (car l) r)
	    (loop (cdr l) r)
	    (loop (cdr l) (cons (car l) r)))
	(begin
	  ;; (display "nodes-union exit " (current-error-port))
	  r))))

;; Test if a new node, based on base nodes, is in l.
;;
;; I: base-nodes, [node]
;;    l, [node]
;; O: #f => nodes does not form a node in l,
;;    otherwise a node of l which has the same representative nodes. 
(define (node-in-list? base-nodes l)
  (let ((node (p-exists? (lambda (x) (and x (nodes-equal? x base-nodes))) l)))
    node))

;; --------------------------------------------------------------------
;; An NFA node has the form ([transition] f s _ no) where:
;;
;; A) Transition is (code node operation) where code is one of 
;;    (.. c1 c2), 
;;    (c c1)
;;    (!.. c1 c2) 
;;    (:: c1 ...) 
;;    (!:: c1 ...). 
;;    where c1, c2 are characters; e is the symbol 'e'
;;          and operation is one of set, snoc, push_empty, push, nop.
;;
;; B) f is #f if it is not a final node, otherwise it may be #t or a tuple
;;   (int symbol func ShortLongPrefer discard?).
;;
;;   It is a tuple if the final node correspond to a token. 
;;
;;   1) int is the position of that token in the list description of tokens,
;;   2) symbol correspond to the token of the TRE
;;   3) func is the name of a procedure,
;;   4) ShortLongPrefer is either the symbol short, long or prefer.
;;   5) discard? is either 'discard or #f
;;
;; C) s is #t iff it is a start node. 
;; D) The fourth element is unused; so that a NFA node is similar 
;;    to a DFA node.
;; E) no is an integer.
;;
;; Some invariants maintained during the construction of the
;; automaton are:
;;
;; - A complete NFA automaton has exactly one start node.
;; - There may be more than one final node, but it cannot happen while
;;   constructing a NFA from other NFAs.
;; - An NFA has more than one final nodes when the final NFA is built,
;;   doing an oring of all tokens.
;; - If a NFA node has more than one out-arc, they are all epsilon-arc.
;;

(define (char-tr? t) 
  (and (pair? t) (eqv? 'c (car t)) (pair? (cdr t)) (char? (cadr t))))

(define (tk-tr? t) (and (pair? t) (eqv? 'tk (car t))))

;; ----------------------------------------------------------------
;; NFA node : #(transitions start? final unused no token-rule)

;; Return node having number no and in nodes.
(define (no-nfa-nodes->nfa-node no nodes)
  (p-exists? (lambda (x) (= no (node-no x))) nodes))

(define (dump-nfa-node x) 
  (display-ln "Nfa no" (node-no x)" Token " (node-TRE-desc x) 
	      " final:" (node-final x)
	      (node-start x))
  (display "Trs ")
  (for-each (lambda (tr) (write (car tr)) (display-ln "-->" (node-no (cadr tr))))
	    (node-trs x)))

;; Add an epsilon transition from n1 to n2.
;; I: n1, n2, NFA nodes. 
;; O: node
(define (a-node-e! n1 n2 op)  (a-node-tr! 'e n1 n2 op))

(define (op->int op) 
  (if (not (pair? op))
      (cdr (assoc op '((snoc . 0) (push . 1) (push_empty . 2) (nop . 99))))
      op))

;; I: x, node
;; O: x
(define (pop-start-state x)
  (if (not (node-start x))
      (error "pop-start-state" "x is not a start state!" x))
  (node-start-set! x #f)
  x)

;; I: x, node.
;; O: x. It must be the start state.
(define (node-start-state? x)  
  (if (not (node-start x) )
      (error "node-start-state" "x is not a start state!." x)
      x))

;; Return #f or the state number of the node start state in S.
;; I: S, [node]
;; O: (u #f int).
(define (nodes->start-node-no S)  
  (let ((n (p-exists? node-start S)))
    (and n (node-no n))))


;; =================================================================
;;
;; The implementation of prefer raises some questions about
;; possible new qualifiers. There are several possibilities
;; for reducing the set of NFA nodes for a DFA node based
;; on the prefer qualifier. Consider a set S of NFA nodes
;;
;; A) If S contains at least one prefer node, even not accepting:
;;    
;;     1) remove all non prefer nodes. 
;;
;;        This is quite drastic and not that useful.
;;
;; B) If S contains at least one accepting node marked as prefer:
;;    
;;    1) Remove all nodes not marked prefer.
;;       
;;       In this case there is no preference based on the position
;;       of TRE definition. This not the adopted case.
;;
;;    2) Remove all nodes not marked prefer and nodes
;;       which are non accepting. 
;;
;;       This is an interesting case but it still does not
;;       take into account the position of the TREs.
;; 
;;    3) Remove all nodes not marked prefer and
;;       nodes marked prefer whose position is above the 
;;       lowest position marked accepting node.
;;
;;       This is the adopted meaning of prefer. This semantic
;;       can clearly be explained:
;;
;;       If a prefer TRE reaches an accepting state, no
;;       non prefer token can be recognized.
;;       If a prefer TRE can reach an accepting state
;;       

;; If there is at least one final node marked as prefer for token t,
;; then return the list of nodes only marked as t. If there
;; are more then one final node marked as prefer, chose the one
;; that has the smallest position number. Otherwise return the
;; original list of nodes. This may be used to force recognition
;; of prefer token as soon as an accepting prefer state is reached.
;; 
;;
;; I: S, [NFA node]
;; O: [NFA node]. Cannot be empty.
;;
(define (reduce-if-final-and-prefer S)
  ;; F is a list of final NFA nodes all marked as prefer.
  ;; O: a NFA node of a TRE with the smallest position.
  (define (smallest-pos F)
    (foldr (lambda (x m) 
	     (if (< (node-pos x) (node-pos m))
		 x
		 m))
	   (car F)
	   (cdr F)))

  ;; F is a list of final NFA nodes.
  ;; Return all nodes marked as prefer.
  (define (nodes->preferNodes F)  (filter node-prefer? F))

  ;; I: S are not necessarily final nodes.
  ;;    n, int, position of TRE.
  ;; O: Return only the nodes which have a lower or equal
  ;;    position than l (higher TRE).
  (define (nodes->higherTRE S n) 
    (filter (lambda (x) (<= (node-pos x) n))  S))

  (let* ((finals  (filter node-final S))
	 (prefers (nodes->preferNodes finals)))
    (if (pair? prefers) ;; at least one final prefer node.
	(let* ((thePreferNode   (smallest-pos prefers))
	       (thePreferNodes  (nodes->higherTRE S (node-pos thePreferNode))))
	  thePreferNodes)
 	S)))

;; Find the final node representative for the set of nodes S. 
;; This set represents a DFA node.
;;
;; I: S, [node]
;; O: (u #f int)
;;    0) If there are no final nodes in S, #f is returned.
;;    
;;    1) If there are no nodes with the 'short' property. The 
;;    	 node with the smallest position number is returned.
;;    
;;    2) If there is only one node with the 'short' property, that node
;;    	 is returned.
;;    
;;    3) If there are more than one node with the 'short' property, the one
;;    	 with the smallest position number is returned.

(define (nodes->final-node-no S)  
  ;; Returns NFA node in F with smallest position.
  (define (smallest-pos F)
    (foldr (lambda (x m) 
	     (if (< (node-pos x) (node-pos m))
		 x
		 m))
	   (car F)
	   (cdr F)))

  (define (nodes->shortNodes F)  (filter node-final-short? F))
  
  (let* ((nodes      (filter node-final S))
	 (shortNodes (nodes->shortNodes nodes)))
    (cond
     ((not (pair? nodes))  #f)
     ((pair? shortNodes) (smallest-pos shortNodes))
     (else               (smallest-pos nodes)))))
  
;; Returns all non-epsilon transition codes out of NFA nodes S.
;; If there is one 'short' final node, returns an empty list.
;;
;; I: S, a set of NFA nodes
;; O: [tr-code]
(define (node-states-trs S)
  ;; (display "node-states-trs..." (current-error-port))
  ;; (flush-output-port (current-error-port))
  (if (p-exists? node-final-short? S)
      '()
      (let* ((trs      (list-append (map node-trs S)))
	     (basicTrs (delete-duplicates (filter (lambda (a) (not (eqv? 'e a)))
						  (map car trs))))
	     (basicTrs2 (break-trs basicTrs))
	     (result    (delete-duplicates basicTrs2)))
	;; (display "exit " (current-error-port))
	result
	)))

;; Break some of the transition codes so that none intersect.
;; This may create some new transition codes.
;;
;; I: [tr-code]
;; O: [tr-code]. No two transition codes intersect.
(define (break-trs trs)
  (let loop ((l trs) (r '()))
    (if (pair? l)
	(let loop2 ((l2 (cdr l)) (x (car l)) (bs '()))
	  (if (pair? l2)
	      (if (and (not (equal? x (car l2)))  (trs-intersect? x (car l2)))
		  (let* ((b (break-intersecting-trs x (car l2))))
		    (loop2 (cdr l2) (car b) (append (cdr b) bs)))
		  (loop2 (cdr l2) x (cons (car l2) bs)))
	      (loop bs (cons x r))))
	(map reduce-tr-code r))))

;; Reduce the size of :: and !:: transition code.
;;
;; I: a transition code; (.. ...), (:: ...), (!:: ...), (!.. ...), (c ...).
;; O: the same transition but express in a different manner.
(define (reduce-tr-code tr-code)
  (if (pair? tr-code)
      (case (car tr-code)
       ((::)
	(let ((range (form-range tr-code)))
	  (cond
	   (range  range)
	   ((> (length (cdr tr-code)) 128)
	    `(!:: ,@(set-complement (cdr tr-code))))
	   (else  tr-code))))
       ((!::)
	(if (> (length (cdr tr-code)) 128)
	    `(:: ,@(set-complement (cdr tr-code)))
	    tr-code))
       (else tr-code))
      tr-code))

;; If possible, transform an :: transition code into a range.
;; I: tr-code, (:: ...)
;; O: #f or (.. ...)
(define (form-range tr-code)
  (let ((cmin (char-min (cdr tr-code)))
	(cmax (char-max (cdr tr-code))))
    (and (= (+ 1 (- (char->integer cmax) (char->integer cmin))) 
	    (length (cdr tr-code)))
	 (list '.. cmin cmax))))

(define (char-min l)
  (and (pair? l)
       (let ((x (char-min (cdr l))))
	 (if (and x (char>? (car l) x))
	     x
	     (car l)))))

(define (char-max l)
  (and (pair? l)
       (let ((x (char-max (cdr l))))
	 (if (and x (char<? (car l) x))
	     x
	     (car l)))))
     
;; Break two transition codes in possibly three transition codes
;; one of which is their intersection. All transition codes
;; are disjoint and cover the two transition codes.
;;
;; I: tr1, tr2, two intersecting transition codes. Not equal.
;; O: [tr-code]. These transition codes do not intersect.
;;          The first transition code of that list is in tr1.
(define (break-intersecting-trs tr1 tr2)
  
  (define (m-.. a b)
    (if (char<? b a)
	'()
	(if (char=? a b)
	    `((c ,a))
	    `((.. ,a ,b)))))
	
  (define (..R a b)
    (if (zero? (char->integer b))
	'()
	(let ((bp  (integer->char (- (char->integer b) 1))))
	  (m-.. a bp))))

  (define (..L a b)
    (if (= 255 (char->integer a))
	'()
	(let ((ap  (integer->char (+ 1 (char->integer a)))))
	  (m-.. ap b))))

  (define (delete-es S l) 
    (foldr (lambda (x r) (delete-e x r)) l S))
  (define (delete-e e l)
    (filter (lambda (x) (not (eqv? e x))) l))

  (if (not (trs-intersect? tr1 tr2))
      (error "break-intersecting-trs" "tr-codes not intersecting! " (list tr1 tr2)))
  (cond
   ((equal? tr1 tr2) `(,tr1))
   ((and (char-tr? tr1) (not (char-tr? tr2)))
    (case (car tr2)
      ((..)    `(,tr1 
		  ,@(..R (cadr tr2) (cadr tr1)) 
		  ,@(..L (cadr tr1) (caddr tr2))))
      ((::)  `(,tr1 
		  ,@(mk-inset (delete-e (cadr tr1) (cdr tr2)))))
      ((!..)   `(,tr1
		  ,@(mk-inset (set-complement (cons (cadr tr1) (range->set tr2))))))
      ((!::) `(,tr1 
		  (!:: ,@(cons (cadr tr1) (cdr tr2)))))
      (else (error "break-intersecting-trs" "unknown case 3" (car tr2)))
      ))
   ((and (not (char-tr? tr1)) (char-tr? tr2)) 
    (break-intersecting-trs tr2 tr1))
   ((and (not (char-tr? tr1)) (not (char-tr? tr2)))
    (case (car tr1)
      ((..)
       (case (car tr2)
	 ((..)   (cond
		   ((ch<=? (cadr tr1) (cadr tr2) (caddr tr1) (caddr tr2))
		    `(,@(..R (cadr tr1) (cadr tr2))
		      ,@(m-.. (cadr tr2) (caddr tr1))
		      ,@(..L (caddr tr1) (caddr tr2))))
		   ((ch<=? (cadr tr2) (cadr tr1) (caddr tr2) (caddr tr1))
		    `(,@(m-.. (cadr tr1) (caddr tr2))
		      ,@(..R (cadr tr2) (cadr tr1))
		      ,@(..L (caddr tr2) (caddr tr1))))
		   ((ch<=? (cadr tr2) (cadr tr1) (caddr tr1) (caddr tr2))
		    `(,tr1
		      ,@(..R (cadr tr2) (cadr tr1))
		      ,@(..L (caddr tr1) (caddr tr2))))
		   ((ch<=? (cadr tr1) (cadr tr2) (caddr tr2) (caddr tr1))
		    `(,tr2
		      ,@(..R (cadr tr1) (cadr tr2))
		      ,@(..L (caddr tr2) (caddr tr1))))))
	 ((:: !.. !::) 
	  (break-intersecting-trs tr2 tr1))
	 (else (error "break-intersecting-trs" "unknown case 4" (car tr2)))
	 ))
      ((::)
       (case (car tr2)
	 ((..) (let ((inter (set-intersection (cdr tr1) (range->set tr2))))
		  `(,@(mk-inset inter)
		    ,@(mk-inset (set-difference (cdr tr1) inter))
		    ,@(mk-inset (set-difference (range->set tr2) inter)))))
	 ((::) (let ((inter (set-intersection (cdr tr1) (cdr tr2))))
		  `(,@(mk-inset inter)
		    ,@(mk-inset (set-difference (cdr tr1) inter))
		    ,@(mk-inset (set-difference (cdr tr2) inter)))))
	 ((!..) 
	  (let ((inter (set-intersection (cdr tr1) 
					 (set-complement (range->set tr2)))))
	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference (cdr tr1) inter))
	      ,@(mk-inset (set-difference (set-complement (range->set tr2))
					  inter)))))
	 ((!::) 
	  (let ((inter (set-intersection (cdr tr1) 
					 (set-complement (cdr tr2)))))
	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference (cdr tr1) inter))
	      ,@(mk-inset (set-difference (set-complement (cdr tr2))
					  inter)))))
	 (else (error "break-intersecting-trs" "unknown case 5" (car tr2)))
	 ))
      ((!..)
       (case (car tr2)
	 ((..) 
	  (let* ((tr1S  (set-complement (range->set tr1)))
		 (tr2S  (range->set tr2))
		 (inter (set-intersection tr1S tr2S)))

	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference tr1S inter))
	      ,@(mk-inset (set-difference tr2S inter)))))
	 ((::) 
	  (let* ((tr1S  (set-complement (range->set tr1)))
		 (tr2S  (cdr tr2))
		 (inter (set-intersection tr1S tr2S)))

	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference tr1S inter))
	      ,@(mk-inset (set-difference tr2S inter)))))
	 ((!..) 
	  (let* ((tr1S  (set-complement (range->set tr1)))
		 (tr2S  (set-complement (range->set tr2)))
		 (inter (set-intersection tr1S tr2S)))

	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference tr1S inter))
	      ,@(mk-inset (set-difference tr2S inter)))))
	 ((!::) 
	  (let* ((tr1S  (set-complement (range->set tr1)))
		 (tr2S  (set-complement (cdr tr2)))
		 (inter (set-intersection tr1S tr2S)))
	    
	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference tr1S inter))
	      ,@(mk-inset (set-difference tr2S inter)))))
	 (else (error "break-intersecting-trs" "unknown case 6" (car tr2)))
	 ))
      ((!::)
       (case (car tr2)
	 ((..) 
	  (let* ((tr1S  (set-complement (cdr tr1)))
		 (tr2S  (range->set tr2))
		 (inter (set-intersection tr1S tr2S)))
	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference tr1S inter))
	      ,@(mk-inset (set-difference tr2S inter)))))
	 ((::) 
	  (let* ((tr1S  (set-complement (cdr tr1)))
		 (tr2S  (cdr tr2))
		 (inter (set-intersection tr1S tr2S)))
	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference tr1S inter))
	      ,@(mk-inset (set-difference tr2S inter)))))
	 ((!..) 
	  (let* ((tr1S  (set-complement (cdr tr1)))
		 (tr2S  (set-complement (range->set tr2)))
		 (inter (set-intersection tr1S tr2S)))
	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference tr1S inter))
	      ,@(mk-inset (set-difference tr2S inter)))))
	 ((!::) 
	  (let* ((tr1S  (set-complement (cdr tr1)))
		 (tr2S  (set-complement (cdr tr2)))
		 (inter (set-intersection tr1S tr2S)))
	    `(,@(mk-inset inter)
	      ,@(mk-inset (set-difference tr1S inter))
	      ,@(mk-inset (set-difference tr2S inter)))))
	 (else (error "break-intersecting-trs" "unknown case 7" (car tr2)))
	 ))
      ))
   (else
    (error "break-intersecting-trs" "unknown transition codes " (list tr1 tr2)))))

;; Create a transition (:: ...) from list of characters l.
(define (mk-inset l)
  (if (null? l)
      '()
      (if (null? (cdr l))
	  `((c ,(car l)))
	  (list (cons ':: l)))))

;; Create a transition (!:: ...) from a list of characters l.
(define (mk-!inset l)
  (if (null? l)
      '()
      (list (cons '!:: l))))

;; Returns all the nodes out of S on transition a. The transition
;; a must be in the transition. It is also verified that a does not
;; intersect, without inclusion, to a transition out of S. If this
;; happens, it is an internal error, probably due to a malfunction of
;; break-intersecting-trs.
;;
;; I: S, [NFA node]
;;    a, (u char (.. char char) (:: char ... char) 
;;               (!.. char char) (!:: char ... char))
;; O: [NFA node]
(define (node-states-tr a S)
  (let* ((trs 
	  (list-append
	   (map (lambda (node) 
		  (filter 
		   (lambda (tr)
		     ;; a in transition code, intersection is not enough!
		     (if (tr-in? a (car tr)) 
			 #t
			 #f
			 ;;(if (trs-intersect? a (car tr))
			 ;;    (error "node-states-tr:" "they intersect! " 
			 ;;	      (list a (car tr)))
			 ;;    #f)
			 ))
		   (node-trs node)))
		S))))
    (delete-duplicates (map cadr trs))))

;; Is transition code a1 in transition code a2?
(define (tr-in? a1 a2)
  (cond
   ((and (char-tr? a1) (char-tr? a2)) ;; (c char)
    (char=? (cadr a1) (cadr a2)))
   ((tk-tr? a1)
    (and (tk-tr? a2) (eqv? (cadr a1) (cadr a2))))
   ((tk-tr? a2) #f) ;; a1 was not a tk.
   ((or (symbol? a1) (symbol? a2))
    (eqv? a1 a2))
   ((and (char-tr? a1) (not (char-tr? a2)))
    (case (car a2)
      ((..)    (ch<=? (cadr a2) (cadr a1) (caddr a2)))
      ((!..)   (not (ch<=? (cadr a2) (cadr a1) (caddr a2))))
      ((::)  (if (memv (cadr a1) (cdr a2)) #t #f))
      ((!::) (not (memv (cadr a1) (cdr a2))))
      (else (error "tr-in?" "unknown transition code " a2))))
   ((and (not (char-tr? a1)) (char-tr? a2))
    (case (car a1)
      ((..)  (ch=? (cadr a1) (caddr a1) (cadr a2)))
      ((!..) (let ((x (set-complement (range->set a1))))
		(and (pair? x) (char=? (car x) (cadr a2)))))
      ((::) (char=? (cadr a1) (cadr a2)))
      ((!::) (let ((x (set-complement (cdr a1))))
		  (and (pair? x) (char=? (car x) (cadr a2)))))
      (else (error "tr-in?" "unknown transition code " a1))))
   ((and (not (char-tr? a1)) (not (char-tr? a2)))
    (case (car a1)
      ((..) (case (car a2)
	       ((..) (ch<=? (cadr a2) (cadr a1) (caddr a1) (caddr a2)))
	       ((!..)   (or (char<? (cadr a2) (cadr a1))
			     (char<? (caddr a1) (caddr a2))))
	       ((::)  (set-in? (range->set a1) (cdr a2) ))
	       ((!::) (set-in? (range->set a1) (set-complement (cdr a2))))
	       (else (error "tr-in?" "unknown transition code " a2))))
      ((!..) (case (car a2)
		((..)  (set-in? (range->set a1)
				 (set-complement (range->set a2))))
		((!..) (set-in? (set-complement (range->set a1))
				 (set-complement (range->set a2))))
		((::) (set-in? (set-complement (range->set a1))
				  (set-complement (cdr a2))))
		((!::) (set-in? (set-complement (cdr a1))
				   (set-complement (cdr a2))))
		(else (error "tr-in?" "unknown transition code " a2)))) 
      ((::) (case (car a2)
		 ((..)   (set-in? (cdr a1) (range->set a2)))
		 ((!..)  (set-in? (cdr a1) (set-complement (range->set a2))))
		 ((::) (set-in? (cdr a1) (cdr a2)))
		 ((!::)(set-in? (cdr a1) (set-complement (cdr a2))))
		 (else (error "tr-in?" "unknown transition code " a2))))
      ((!::) (case (car a2)
		  ((..)   (set-in? (set-complement (cdr a1)) (range->set a2)))
		  ((!..)  (set-in? (set-complement (cdr a1)) 
				    (set-complement (range->set a2))))
		  ((::) (set-in? (set-complement (cdr a1)) (cdr a2)))
		  ((!::)(set-in? (set-complement (cdr a1)) 
				    (set-complement (cdr a2))))
		  (else (error "tr-in?" "unknown transition code " a2))))
      (else (error "tr-in?" "Unknown transition code " a1))))
   (else (error "tr-in?" "Unknown transition codes " (list a1 a2)))))
 

;; Do transition codes a1 and a2 intersect?
;; 
(define (trs-intersect? a1 a2)
  (cond
   ((and (char-tr? a1) (char-tr? a2))
    (char=? (cadr a1) (cadr a2)))
   ((tk-tr? a1) 
    (and (tk-tr? a2) (eqv? (cadr a1) (cadr a2))))
   ((tk-tr? a2)   #f) ;; a1 was not a tk.
   ((or (symbol? a1) (symbol? a2))
    (eqv? a1 a2))
   ((or (not (pair? a1)) (not (pair? a2)))
    (error "trs-intersect?" "unknown transitions: " (list a1 a2)))
   ((and (char-tr? a1) (not (char-tr? a2)))
    (case (car a2)
      ((..)    (ch<=? (cadr a2) (cadr a1) (caddr a2)))
      ((!..)   (not (and (ch<=? (cadr a2) (cadr a1))
			  (ch<=? (cadr a1) (caddr a2)))))
      ((::)  (memv (cadr a1) (cdr a2)))
      ((!::) (not (memv (cadr a1) (cdr a2))))
      (else (error "trs-intersect?" "unknown case 0 " a2))))
   ((and (not (char-tr? a1)) (char-tr? a2))
    (trs-intersect? a2 a1))
   ((and (not (char-tr? a1)) (not (char-tr? a2)))
    (case (car a1)
      ((..) (case (car a2)
	       ((..) (or (and (ch<=? (cadr a1) (cadr a2))
			       (ch<=? (cadr a2) (caddr a1)))
			  (and (ch<=? (cadr a2) (cadr a1))
			       (ch<=? (cadr a1) (caddr a2)))))
	       ((!..)   (or (char<? (cadr a1) (cadr a2))
			     (char<? (caddr a2) (caddr a1))))
	       ((::)  (set-intersect? (range->set a1) (cdr a2)))
	       ((!::) (set-intersect? (set-complement (cdr a2))
					 (range->set a1)))
	       (else (error "trs-intersect?" "unknown case 1" a2))
	       ))
      ((!..) (case (car a2)
		((..)  (trs-intersect? a2 a1))
		((!..) (set-intersect? (set-complement (range->set a1))
					(set-complement (range->set a2))))
		((::) (set-intersect? (set-complement (range->set a1))
					 (cdr a2)))
		((!::) (set-intersect? (set-complement (range->set a1))
					  (set-complement (cdr a2))))
		 (else (error "trs-intersect?" "unknown case 2 " a2))
		))
      ((::) (case (car a2)
		 ((..)   (trs-intersect? a2 a1))
		 ((!..)  (trs-intersect? a2 a1))
		 ((::) (set-intersect? (cdr a1) (cdr a2)))
		 ((!::) (set-intersect? (cdr a1) (set-complement (cdr a2))))
		 (else (error "trs-intersect?" "unknown case 3 " a2))
		 ))
      ((!::) (case (car a2)
		  ((..)    (trs-intersect? a2 a1))
		  ((!..)   (trs-intersect? a2 a1))
		  ((::)  (trs-intersect? a2 a1) )
		  ((!::) (set-intersect? (set-complement (cdr a1))
					    (set-complement (cdr a2))))
		  (else (error "trs-intersect?" "unknown case 4 " a2))
		  ))
      (else (error "trs-intersect?" "Unknown transition " a1))))
   (else (error "trs-intersect?" "Unknown transitions " (list a1 a2)))))

;; Returns the epsilon closure from the NFA nodes S.
;;
;; I: S, [NFA nodes]
;; O: [NFA nodes]. (It includes S).
(define (e-closure S)
  (define (remove-nodes-with-only-epsilon-trs nodes)
    (filter (lambda (node)
	      (or (node-final node)
		  (node-start node)
		  (p-exists? (lambda (tr) (not (eqv? 'e (car tr))))
			     (node-trs node))))
	    nodes))

  ;; S: nodes to be visited.
  (define (e-closureR S visited)
    (if (null? S)
	visited
	(let* ((nodes   (delete-duplicates
			 (map cadr
			      (list-append
			       (map (lambda (v)
				      (filter (lambda (tr) (eqv? 'e (car tr))) 
					      (node-trs v))) 
				    S)))))
	       (nodes-not-seen (set-difference nodes visited)))
	  (e-closureR nodes-not-seen (append nodes-not-seen visited)))))
  (let* ((all-visited-nodes (e-closureR S S)))
    (remove-nodes-with-only-epsilon-trs all-visited-nodes)))

  
;; Return the final state of the FA and mark it as non-final.  There
;; should be only one such state, otherwise it is an internal error.
;;
;; I: x, FA node.
;; O: FA node. 
(define (pop-final-state x)
  (let* ((final-states (search-final-states x)))
    (if (not (pair? final-states))
	(error "pop-final-state" "No final state!" x))
    (if (pair? (cdr final-states))
	(error "pop-final-state" "there are more than one final state" x))
    (node-final-set! (car final-states) #f)
    (car final-states)))

;; Return the final states of the DFA.
;;
;; I: x, DFA node.
;; O: [DFA node]
(define (search-final-states x)
  (let* ((reachable-states (search-reachable-states x)))
    (filter node-final reachable-states)))

;; O: ([DFA-final-nodes] [DFA-non-final-nodes])
(define (search-final-and-non-final-states x)
  (let* ((states     (search-reachable-states x))
	 (finals     (filter node-final states))
	 (non-finals (filter (lambda (x) (not (node-final x))) states)))
    (list finals non-finals)))

;; Return all reachable nodes from x.
;; O: [DFA node]
(define (search-reachable-states x)
  (let loop ((l (list x)) (r '()))
    (if (pair? l)
	(let* ((node (car l)))
	  (if (memq node r) ;; do not visit twice.
	      (loop (cdr l) r)
	      (loop (append (map cadr (node-trs node)) (cdr l))
		    (cons node r)
		    )))
	r
	)))

;; ====================================================================
;;
;; Transformation from a DFA to a vector for the lexer driver.

;; I: n, the DFA starting node.
;; O: vector of vectors #([(tr no)] start? final?)
;;                         start?: #f or integer of start NFA node.
;;                         final?: #f or integer of final NFA node.
(define (dfa->vector n)
  (let* ((r          '())
	 (max-node-no 0))

    (define (path-ops->drvr-ops l-paths)
      (map (lambda (path)
	     (list (car path) (cadr path)
		   (map op->int (caddr path))))
	   l-paths))

    (define (tr-code->drvr-tr-code tr-code)
      (if (pair? tr-code) 
	  (case (car tr-code)
	    ((:: !::) (cons (car tr-code) (map char->integer (cdr tr-code))))
	    ((.. !..) (list (car tr-code) (char->integer (cadr tr-code))
			      (char->integer (caddr tr-code))))
	    ((c) (char->integer (cadr tr-code)))
	    (else 
	     tr-code))
	  tr-code))
    
    ;; I: n, node
    ;;    S, already visited nodes.
    (define (dfa-visit n S)
      (if (> (node-no n) max-node-no) (set! max-node-no (node-no n)))
      (if (not (memq n S))
	  (let* ((final-tokens (delete-duplicates 
				(map node-tk (node-base-nodes n))))
		 (transitions 
		  (map (lambda (tr)
			 (let ((tr-code (car tr))
			       (to-node (cadr tr))
			       (l-paths (caddr tr)))
			   (list (tr-code->drvr-tr-code tr-code)
				 (node-no to-node)
				 (path-ops->drvr-ops l-paths)
				 )))
		       (node-trs n))))
	    (set! r (cons 
		     (cons (node-no n)
			   (m-va transitions
				 (and (node-start n) 
				      (path-ops->drvr-ops (node-start n)))
				 (and (node-final n)
				      ;; (int symbol), symbol is the token id
				      `(,(node-no (node-final n))
					,@(cdr (node-TRE-desc (node-final n)))
					))
				 final-tokens
				 ))
		     r
		     ))

	    (foldr (lambda (x l) (dfa-visit x l))
		   (cons n S)
		   (map cadr (node-trs n))))
	  S))
    (if (not (node-start n))
	(error "dfa->vector" "node n is not a starting node." n))
    (dfa-visit n '())
    (let* ((v (make-vector (+ max-node-no 1) #f)))
      (for-each
       ;; x: (int . vector)
       (lambda (x) (vector-set! v (car x) (cdr x)))
       r)
      v)))

;; A va node is essentially a serialized DFA node.
;;
;; n: DFA node used to construct this va node.
;; start?: #t or #f.
;; final?: #f or (int ...) where int is a NFA node number. For ... see
;;            the cdr of node-TRE-desc.
;; tokens: list of recognized tokens (rule tokens) that this
;;         node may lead to: there is a path from that node
;;         to a final node marked with one of the tokens.

(define (m-va trs start? final? tokens)  
  (vector trs start? final? tokens))

(define (va-trs v)               (vector-ref v 0))
(define (va-start? v)            (vector-ref v 1))
(define (va-final  v)            (vector-ref v 2))
(define (va-tokens v)            (vector-ref v 3))
  
;; ---------------------------
;; Library of functions.


;; Return an element of l satisfying p; or #f.
(define (p-exists? p l)
  (if (pair? l)
      (or (if (p (car l)) (car l) #f) (p-exists? p (cdr l)))
      #f))

;; I: p, a function of n arguments.
;;    ls, a list of n>0 lists of same length.
;; O: #t iff for all (p a1 ... an). 
(define (for-all p . ls)
  (if (pair? (car ls))
      (and (apply p (map car ls)) (apply for-all p (map cdr ls)))
      #t))

(define every for-all)

(define (list-append l)
  (foldr (lambda (e L) (append e L)) '() l))

(define (foldr f a l)
  (if (pair? l)
      (f (car l) (foldr f a (cdr l)))
      a))

(define (filter p l)
  (let loop ((l l) (r '()))
    (if (pair? l)
	(if (p (car l))
	    (loop (cdr l) (cons (car l) r))
	    (loop (cdr l) r))
	(reverse r))))

(define (display-ln . l) (for-each display l)  (newline))
(define (display-l  . l) (for-each display l))
(define (write-l    . l) (for-each write l))
(define (write-ln   . l) (for-each write l) (newline))

(define (set-complement l)
  (set-difference (range->set `(.. ,(integer->char 0) ,(integer->char 255)))
		  l))
  
;; I: a, (.. c1 c2), where c1 and c2 are characters.
;; O: [char]
(define (range->set a)
  (let loop ((i (char->integer (cadr a))) (r '()))
    (if (<= i (char->integer (caddr a)))
	(loop (+ i 1) (cons (integer->char i) r))
	r)))

(define (set-difference e1 e2)
  (let loop ((l e1) (r '()))
    (if (pair? l)
	(if (memq (car l) e2)
	    (loop (cdr l) r)
	    (loop (cdr l) (cons (car l) r)))
	r)))

(define (delete-duplicates l)
  (let loop ((l l) (r '()))
    (if (pair? l)
	(if (fast-member (car l) r)
	    (loop (cdr l) r)
	    (loop (cdr l) (cons (car l) r)))
	r
	)))

(define (fast-member e l)
  (if (vector? e)
      (memq e l)
      (member e l)))

(define (set-intersect? e1 e2)
  (or (and (pair? e1)
	   (or (memv (car e1) e2) (set-intersect? (cdr e1) e2)))))

(define (set-in? e1 e2)
  (or (null? e1)
      (and (pair? e1)
	   (memv (car e1) e2)
	   (set-in? (cdr e1) e2))))

(define (set-intersection l1 l2)
  (let loop ((l l1) (r '()))
    (if (pair? l)
	(if (memv (car l) l2)
	    (loop (cdr l) (cons (car l) r))
	    (loop (cdr l) r))
	r)))

(define (set-union l1 l2)
  (let loop ((l (append l1 l2)) (r '()))
    (if (pair? l)
	(if (memv (car l) r)
	    (loop (cdr l) r)
	    (loop (cdr l) (cons (car l) r)))
	r)))

(define (list-one-element? l)
  (and (pair? l) (null? (cdr l))))

(define (set-equal? s1 s2)
  (and (= (length s1) (length s2))
       (for-all (lambda (x) (fast-member x s1)) s2)))

;; I: l, [str]
;; O: str
(define (list-string-append l)
  (if (pair? l)
      (string-append (car l) (list-string-append (cdr l)))
      ""
      ))

;; I: declare, [(parm value)]
;; O: sym.
(define (get-parm parm-name declare)
  (let ((v (search-parm parm-name declare)))
    (if (not v)	(err "No parameter " parm-name " in declaration " declare "."))
    (if (not (symbol? v))
	(err "A parameter value must be a symbol, but it is " v))
    v))

(define (search-parm k l)
  (let loop ((l l))
    (if (and (pair? l) (pair? (car l))
	     (pair? (cdar l)))
	(if (eqv? (caar l) k)
	    (cadr (car l))
	    (loop (cdr l)))
	#f)))
	    
(define (for-eachi p l n)
  (let loop ((l l) (i n))
    (if (pair? l)
	(begin
	  (p (car l) i)
	  (loop (cdr l) (+ i 1))))))

(define (mapi p l)
  (let loop ((l l) (i 0) (r '()))
    (if (pair? l)
	(loop (cdr l) (+ i 1) (cons (p (car l) i) r))
	(reverse r))))

;; Apply p to every element of l from left to right. Feed every result to p as in foldl.
;; Similar to scan with a position parameter.
(define (mapi-foldl p s l)
  (let loop ((l l) (r '()) (last-res s) (i 1))
    (if (pair? l)
	(let ((res (p (car l) last-res i)))
	  (loop (cdr l) (cons res r) res (+ i 1)))
	(reverse r))))
	  
(define (my-list-ref l n) (and (< n (length l)) (list-ref l n)))

(define (ch<=? a b . l)
  (and (char<=? a b)
       (let loop ((l l) (p b))
	 (or (not (pair? l))
	     (and (char<=? p (car l)) 
		  (loop (cdr l) (car l)))))))

(define (ch=? a b . l)
  (and (char=? a b)
       (let loop ((l l) (p b))
	 (or (not (pair? l))
	     (and (char=? p (car l)) 
		  (loop (cdr l) (car l)))))))


(define (warning . l)
  (newline)(display "Warning: ")
  (for-each display l)
  (newline))
	     
(define (err . l)
  (newline)
  (for-each display l)
  (newline)
  (error "err" "Stopping execution." #t))


