;; $Id: regregDrvr.scm,v 1.6 2003/05/28 01:12:22 latendre Exp $
;;
;; The driver for all RegReg parser instances.

;; Copyright (c) 2002, Mario Latendresse
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;;
;;     * Redistributions of source code must retain the above copyright
;; notice, this list of conditions and the following disclaimer.
;; 
;;     * Redistributions in binary form must reproduce the above
;; copyright notice, this list of conditions and the following disclaimer
;; in the documentation and/or other materials provided with the
;; distribution.
;; 
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;; FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
;; COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;; INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
;; (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
;; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
;; STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
;; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
;; OF THE POSSIBILITY OF SUCH DAMAGE.


;; Put #t instead of #f (after if) to get an execution trace of your parser.

(define-macro (drvr:debug-trace . l)
  (if #f `(display-l ,@l) #f))

;; This function can be used to parse a string.
;;
;; I: s, str.
;;    na, a vector as generated from regreg.
;; O: the tagged parse tree.
(define (parse-string s na)
  (with-input-from-string s
    (lambda ()
      (let* ((lex  (make-parser read-char na "")))
	(lex)))))

;; Construct the full parser based on DFA vectors.
;;
;; I: readc, a function of zero argument to read the input characters.
;;    dfa-vectors, as generated by regreg.
;;    buf-primer, string to initialize basic lexer buffer.
(define (make-parser readc dfa-vectors buf-primer)
  ;; i is the level of the parser.
  (foldli (lambda (i r x) (make-lexer-token r (cadr x) (car x) i))
	  (make-lexer readc (cadar dfa-vectors) (caar dfa-vectors) buf-primer)
	  (cdr dfa-vectors)
	  2))

;; ==============================================================
;; The actual interface between the user code and the parse tree

;; Returns a string or #f.
(define (gstks key t)
  (tree->string (gstree key t)))

;; Get a subtree from t having key k. 
;; I: k is a symbol; t is the result, or sub-result, of the parser.
;; O: the `value' (the cadr) of k.
(define (gstree k t)
  (let ((r  (and (pair? t)
		 (if (eqv? k (car t))
		     (cadr t)
		     (or (and (pair? (car t)) (gstree k (car t)))
			 (p-first-one-of (lambda (x) (gstree k x)) (cdr t)))))))
    (if (and (pair? r) (null? (cdr r)) (string? (car r)))
	(car r)
	r)))

;; Return a list of all subtrees of t having key k. The subtrees are disjoint.

(define (gstree-all k t)
  (if (pair? t)
      (if (eqv? k (car t))
	  (if (and (pair? (cadr t)) (null? (cdadr t)) (string? (caadr t)))
	      (list (caadr t))
	      (list (cadr t)))
	  (if (pair? (car t))
	      (append (gstree-all k (car t)) (gstree-all k (cdr t)))
	      (gstree-all k (cdr t))))
      '()))

;; The list l form a tree for which the leaves are strings. This
;; function concatenates all these strings into one. l may be a string.
;;
(define (tree->string l)
  (and l
       (cond
	((pair? l)   (drvr:list-string-append (map tree->string l)))
	((string? l) l)
	(else ""))))

;; End of the interface 
;; =============================================================


;; =======================================================================

;; The following functions should not be used directly.

;;
;; I: read-token, () -> (sym result)
;;    va-dfa as produced by dfa->vector.
;;    level, int.
;; O: 
(define (make-lexer-token read-token va-dfa error-func level)
  (let* ((max-buf-length 1000000)              ;; 1000K tokens, maximum.
	 (buf-length     200)                  ;; initial length of buf
	 (buf         (make-vector buf-length #f)) ;; vector of (sym result)
	 ;; Three pointers in the buffer. 
	 (iend         0) ;; index to store in buf
	 (icur        #f) ;; index to read current (sym result)
	 (ibeg         0) ;; first (sym result) in buf
	 ;; The current reading function.
	 (cur-read     #f)
	 (eof-reached  #f)
	 )

    ;; When reading from buffer, there is still at least one space
    ;; left to store into it.

    ;; There is a token in buf to read; ibeg <= icur < iend.
    ;; O: (sym result)
    (define (read-from-buffer1)
      (let ((tk  (vector-ref buf icur)))
	(set! icur (+ 1 icur))
	(if (= icur iend) 
	    (begin 
	      (set! cur-read read-in1)
	      (set! icur #f)))
	tk))
  
    ;; There is a token in buf to read; icur < iend < ibeg. 
    ;; O: (sym result)
    (define (read-from-buffer2)
      (let ((tk (vector-ref buf icur)))
	(set! icur (+ 1 icur))
	(if (= icur iend) 
	    (begin
	      ;; No longer reading from the buffer.
	      (set! icur #f)
	      (set! cur-read read-in2)))
	tk))
    
    ;; There is a token in buf to read; iend < ibeg <= icur. 
    ;; O: (sym result)
    (define (read-from-buffer3)
      (let ((tk (vector-ref buf icur)))
	(set! icur (+ 1 icur))
	(if (= icur buf-length) 
	    (if (= iend 0) ;; a rare case
		(begin  ;; No longer reading from the buffer.
		  (set! icur #f)
		  (set! cur-read read-in2))
		(begin 
		  (set! icur 0)
		  ;; icur < iend < ibeg, so cur-readc should be changed.
		  (set! cur-read read-from-buffer2))))
	tk))

    ;; This effectively uses the read-token procedure.  We may assume
    ;; that there is space in buf to store a token and
    ;; that ibeg =< iend.
    ;; O: (sym result)
    (define (read-in1)
      (let* ((tk (read-token)))
	(if (or (eqv? (car  tk) '<error>) (eqv? (car tk) '<eof>))
	    tk
	    (begin
	      (vector-set! buf iend tk) 
	      (set! iend (+ 1 iend))
	      (if (= iend buf-length)
		  (if (= ibeg 0) ;; a rare case
		      (enlarge-buffer!)
		      (begin
			(set! iend 0)
			(set! cur-read read-in2))))
	      tk))))

    ;; The same condition as read-in1 applies but iend < ibeg.
    ;; O: (sym result)
    (define (read-in2)
      (let* ((tk (read-token)))
	(if (or (eqv? (car  tk) '<error>) (eqv? (car tk) '<eof>))
	    tk
	    (begin
	      (vector-set! buf iend tk)
	      (set! iend (+ 1 iend))
	      (if (= iend ibeg) (enlarge-buffer!))
	      tk))))

    ;; O: void.
    (define (enlarge-buffer!)
      (if (> (* 2 buf-length) max-buf-length)
	  (error "Enlarge-buffer!" "maximum length reached, (length . max): "
		 (cons buf-length max-buf-length)))

      (let* ((new-buf-length   (* 2 buf-length))
	     (new-buf          (make-vector new-buf-length))
	     (length-to-copy   (if (<= iend ibeg)
				   (- (vector-length buf) (- iend ibeg))
				   (- iend ibeg))))
	;; buf is full: copy from buf to new-buf.
	(if (= 0 length-to-copy)
	    (error "enlager-buffer!" "buffer is empty. " 
		   (list iend ibeg (vector-length buf))))
	
	(let loop ((i ibeg) (j 0))
	  (if (< j length-to-copy)
	      (if (< i buf-length)
		  (begin 
		    (vector-set! new-buf j (vector-ref buf i))
		    (loop (+ i 1) (+ j 1)))
		  (loop 0 j))))
	
	(set! buf  new-buf)  ;; forget buf
	(set! buf-length  new-buf-length)
	(set! ibeg 0)
	(set! iend length-to-copy)
	(set! cur-read read-in1)))

    ;; O: void
    (define (reset-buffer!)
      (set! iend 0)
      (set! ibeg 0)
      (set! icur #f)
      (set! cur-read read-in1))

    (define (buffer-contents)
      (if (< ibeg iend)
	  (let loop ((i ibeg) (r '()))
	    (if (< i iend)
		(loop (+ i 1) (cons (vector-ref buf i) r))
		(reverse r)))
	  (let loop ((i ibeg) (r '()))
	    (if (= i buf-length)
		(loop 0 r)
		(if (not (= i iend))
		    (loop (+ i 1) (cons (vector-ref buf i) r))
		    (reverse r))))))

    ;; -- Begin make-lexer-token --   
    (if (not (va-start? (vector-ref va-dfa 1)))
	(error "\nmake-lexer-token" "The first node is not a start node." va-dfa))
    (if (= buf-length 0)
	(error "\nmake-lexer-token " "The initial buffer length is zero!" buf-length)) 
    (set! cur-read read-in1)

    ;; The returned thunk to read tokens.
    (lambda ()
      (let ((last-final  #f) ;; see function possible-final-node.
            (no-token-used #f)
	    (last-ptr-tr #f) ;; points the last-final node in list trace-trs.
	    (trace-trs  '())) ;; All nodes that the parser went through.
	
	;; At this level a and b are not characters.
	;; This is inefficient but the length of b is usually short.
	(define (snoc a b)  (append b (list a)))

	;; Return the list of rule tokens from the last taken transition.
	(define (get-rule-tokens)
	  (if (pair? trace-trs)
	      (let* ((tr  (car trace-trs))
		     (no  (or (cadr tr) (and (pair? (cdr trace-trs))
					     (cadr (cadr trace-trs))))))
		(if no
		    (va-tokens (vector-ref va-dfa no))
		    '()))
	      '()))

	(define (discard-tokens)
	  (if (not icur)
	      ;; Tokens are not read from buffer. Therefore, all
	      ;; stored tokens can be discarded.
	      (begin
		(if (= ibeg iend)
		    (if no-token-used
			(error "discard-tokens"
			       "RegReg lexer terminated: recognized empty string of tokens twice."
			       "1")
			(set! no-token-used #t))
		    (set! no-token-used #f))
		;; Just reset the buffer, forget its content.
		(set! ibeg 0)
		(set! iend 0))
	      (begin
		;; Tokens were read from the buffer. icur points the next
		;; char to read.
		(if (= icur ibeg)
		    (if no-token-used
			(error "discard-chars"
			       "RegReg lexer terminated: recognized empty string twice."
			       "2")
			(set! no-token-used #t))
		    (set! no-token-used #f))
		
		(set! ibeg icur)
		;; Reset icur and cur-read to appropriate values.
		(if (= ibeg iend)
		    (begin ;; There are no tokens to read from buffer.
		      (set! icur #f)
		      (set! cur-read read-in1))
		    (begin ;; There are tokens to read from buffer.
		      ;; icur = ibeg
		      (cond
		       ((< ibeg iend)  (set! cur-read read-from-buffer1))
		       (else (set! cur-read read-from-buffer3))))))))


	;; I: ops, [op].
	;; O: 
	(define (apply-grammar-ops ops)
	  ;; ibeg is the first (sym result) in buffer.
	  (let loop ((l ops) (i ibeg) (stack '()))
	    (if (pair? l)
		(if (pair? (car l))
		    (if (eqv? 'set (caar l))
			(loop (cdr l) i 
			      ;; bind an id with the top value on stack.
			      (cons (list (cadr (car l)) (car stack))
				    (cdr stack)))
			(error "apply-grammar-ops" "unknown operation " (car l)))
		    (case (car l)
		      ((2) (loop (cdr l) i (cons '() stack)))
		      ((1)
		       (loop (cdr l) 
			     (if (= (+ i 1) buf-length) 0 (+ i 1)) 
			     (cons (vector-ref buf i) stack)
			     ))
		      ((0)
		       (let* ((a (car stack))
			      (b (cadr stack)))
			 (loop (cdr l) i (cons (snoc a b) (cddr stack)))))
		      ;; push snoc
		      ((3)
		       (let* ((b (car stack)))
			 (loop (cdr l) 
			       (if (= (+ i 1) buf-length) 0 (+ i 1)) 
			       (cons (snoc (vector-ref buf i) b) (cdr stack)))))
		      (else 
		       (error "apply-grammar-ops" "unknown operation " (car l)))))
		(begin ;; End of application, adjust the current read.
		  ;; 0 <= i < buf-length
		  (set! ibeg i)
		  ;; Reset icur and cur-read to appropriate values.
		  (if (= ibeg iend)
		      (begin 
			(set! icur #f)
			(set! cur-read read-in1))
		      (begin 
			(set! icur i)
			;; icur = ibeg
			(cond
			 ((< ibeg iend)  (set! cur-read read-from-buffer1))
			 (else (set! cur-read read-from-buffer3)))))
		  ;; Return result of the lexer
		  (car stack)))))

	;;
	;; I: 
	(define (apply-all-token-ops)
	  (cond
	   (last-ptr-tr
	    (let* ((trs    last-ptr-tr)
		   (discard? (list-ref last-final 4))
		   (ops    (and (not discard?)
				(build-ops-path va-dfa (car last-final) trs)))
		   (r      (or (and (not discard?) (apply-grammar-ops ops))
			       '()))
		   (id     (cadr last-final))
		   (r2     (list id r))
		   (func   (list-ref last-final 2))
		   (result (if func
			       ((list-ref last-final 2) r2)
			       r2)))
	      (if discard? (discard-tokens))
	      (drvr:debug-trace "\n\nLexer "level" result: "result)
	      result))
	   ((and eof-reached (null? (cdr trace-trs)))
	    ;; Eof was reached and no token at the previous level was recogized.
	    '(<eof>))
	   (else
	    ;; Error: No grammar rule has been recognized.
	    (let* ((r      (buffer-contents))
		   (tokens (get-rule-tokens)))
	      (drvr:debug-trace "\nNo grammar rule recognized at level "level"!")
	      (drvr:debug-trace "\nCurrent list of recognized tokens: " r)
	      (drvr:debug-trace "\nCurrent token states: " tokens)
	      (reset-buffer!)
	      (if error-func
		  (error-func tokens r)
		  `(<error> ,tokens ,r)
		  )))))

	;; I: tr, a transition from the grammar DFA, 
	;;        that is (code-tr next-node-no (no no [op]))
	(define (trace-tr! tr)  (set! trace-trs (cons tr trace-trs)))

	;; I: (int sym proc type discard?). 
	;;    The int is a NFA node number. 
	;;    The symbol is the identifier of the grammar rule.
	;;    The proc is #f or the processing function to call with the result.
	;;    The type is long or short. (Note: a final short node does not 
	;;                                have any outgoing edge.)
        ;;    discard? = discard => do not build the parse tree.
	(define (possible-final-node final)
	  (set! last-ptr-tr trace-trs)
	  (set! last-final final))

	;; I: tk, (token result).
	;;    dn, an element of the grammar va vector.
	(define (apply-tr tk dn)
	  (drvr:debug-trace "\nPossible transitions " (map cadar (va-trs dn)))
	  (let loop ((trs (va-trs dn)))
	    (if (pair? trs)
		(if (eqv? (car tk) (cadaar trs))
		    (begin
		      (trace-tr! (car trs))
		      ;; return index of new DFA state.
		      (cadr (car trs)))
		    (loop (cdr trs)))
		#f)))
	
	(define (parse-input)	
	  ;; tk: (token result)
	  (let loop ((tk (cur-read)) (cn (vector-ref va-dfa 1)))
	    (drvr:debug-trace "\n\nLexer "level", token received "(car tk) ", active rules " (get-rule-tokens))
	    (cond
	     ((eqv? (car tk) '<error>)
	      (let ((r      (buffer-contents))
		    (tokens (get-rule-tokens)))
		(if error-func 
		    (begin
		      (reset-buffer!)
		      (error-func (cons tokens (list (cadr tk)))
				  (list r (caddr tk))) )
		    (begin
		      (reset-buffer!)
		      `(<error> ,(cons tokens (list (cadr tk)))
				,(list r (caddr tk)))))))
	     ((eqv? (car tk) '<eof>)
	      (set! eof-reached #t)
	      (apply-all-token-ops))
	     (else
	      (let* ((i   (apply-tr tk cn))
		     (cnp (and i (vector-ref va-dfa i))))
		(if cnp
		    (begin
		      (if (va-final cnp) (possible-final-node (va-final cnp)))
		      (if (pair? (va-trs cnp)) 
			  (loop (cur-read) cnp)
			  (apply-all-token-ops)))
		    ;; No transition is applicable
		    (apply-all-token-ops)
		    ))))))

	(drvr:debug-trace "\nLexer at level " level)
	(if eof-reached
	    ;; This is eof before reading anything: the buffer is empty and
	    ;; there is no trace of transitions.
	    '(<eof>)
	    (let ((first-node (vector-ref va-dfa 1)))
	      ;; Trace the start node when entering the automaton.
	      (trace-tr! (list #f 1 (va-start? first-node)))
	      (if (va-final first-node) 
		  (possible-final-node (va-final first-node)))
	      (parse-input)
	      ))))))

;; ======================================================================
;;
;; make-lexer and make-lexer-token are similar. They are made separate
;; for efficiency.  The lexer generated by make-lexer necessarily
;; treats characters as readc reads characters.


;; I: readc, a function to read one character.
;;    dfa-vector, as build by dfa->vector.
;;    error-func, (str)-> . This function receives a string, which is the input read
;;                          up to the point that no transition could be applied.
;;    buf-primer, a string. It can be used to initialize the buffer
;;
;; O: proc: () -> (token result)
(define (make-lexer readc dfa-vector error-func buf-primer)
  (let* ((max-buf-length   200000)              ;; applies only to buf
	 (buf-length       500)                 ;; initial length of buf
	 (buf         (make-string buf-length)) ;; buf stores all characters read
	 ;; Three pointers in the buffer. 
	 (iend         0) ;; index to store in buf
	 (icur        #f) ;; index to read current char
	 (ibeg         0) ;; beginning of token in buf
	 ;; The current reading function.
	 (cur-readc    #f)
	 (eof-reached  #f)
	 (no-char-used #f) ;; if #t, no character was used from buffer or
	                   ;; input. This is used to detect possible
	                   ;; infinite looping and abort the lexer.
	 )

    ;; Should only be called at the creation of the lexer, not
    ;; to reset the buffer buf.
    ;; O: void
    (define (init-buf)
      (let ((n (string-length buf-primer)))
	(if (= 0 n)
	    (set! cur-readc read-in1)
	    (begin
	      (if (>= n max-buf-length)
		  (error "make-lexer" "init string is not smaller than max-buf-length, buf-primer length: " n))
	      (if (>= n buf-length)
		  (begin
		    (set! buf (make-string (+ n 1)))
		    (set! buf-length (+ n 1))))
	      (do ((i 0 (+ i 1)))
		  ((= i n))
		(string-set! buf i (string-ref buf-primer i)))
	      (set! iend n) ;; n>0
	      (set! icur 0) ;; read from buffer at icur
	      (set! cur-readc read-from-buffer1)))))

    ;; Some invariants:
    ;;
    ;; 0 <= iend < buf-length
    ;; iend = ibeg = 0 => no character in the buffer.
    ;; icur = #f <=> next char not read from buffer

    ;; The buffer is never full after read-in1 and read-in2.

    ;; There are several reading routines to keep track of some state
    ;; thus to increase performance.

    ;; This effectively uses the readc procedure.  We may assume
    ;; that there is space in buf to store the character and
    ;; that ibeg =< iend.
    ;;
    ;; O: char
    (define (read-in1)
      (let* ((c (readc)))
	(if (eof-object? c)
	    c
	    (begin
	      (string-set! buf iend c)
	      (set! iend (+ 1 iend))
	      (if (= iend buf-length)
		  (if (= ibeg 0)
		      (enlarge-buffer!)
		      (begin
			(set! iend 0)
			(set! cur-readc read-in2))))
	      c))))

    ;; This effectively uses the readc procedure.  We may assume
    ;; that there is space in buf to store a character and
    ;; that iend < ibeg.
    ;;
    ;; O: char
    (define (read-in2)
      (let* ((c (readc)))
	(if (eof-object? c)
	    c
	    (begin
	      (string-set! buf iend c)
	      (set! iend (+ 1 iend))
	      (if (= iend ibeg) (enlarge-buffer!))
	      c))))

    ;; When reading from buffer there is space to store at least one char at iend.

    ;; There is a char in buf to read; ibeg <= icur < iend.
    ;;
    ;; O: char
    (define (read-from-buffer1)
      (let ((c  (string-ref buf icur)))
	(set! icur (+ 1 icur))
	(if (= icur iend) 
	    (begin 
	      ;; No longer reading from the buffer.
	      (set! cur-readc read-in1)
	      (set! icur #f)))
	c))
  
    ;; There is a char in buf to read; icur < iend < ibeg. 
    ;;
    ;; O: char
    (define (read-from-buffer2)
      (let ((c (string-ref buf icur)))
	(set! icur (+ 1 icur))
	(if (= icur iend) 
	    (begin
	      ;; No longer reading from the buffer.
	      (set! icur #f)
	      (set! cur-readc read-in2)))
	c))
    
    ;; There is a char in buf to read; iend < ibeg <= icur. 
    ;;
    ;; O: char
    (define (read-from-buffer3)
      (let ((c (string-ref buf icur)))
	(set! icur (+ 1 icur))
	(if (= icur buf-length) 
	    (if (= iend 0) ;; a rare case
		(begin 
		  ;; No longer reading from the buffer.
		  (set! icur #f)
		  (set! cur-readc read-in2))
		(begin 
		  (set! icur 0)
		  ;; icur < iend < ibeg, so cur-readc should be changed.
		  (set! cur-readc read-from-buffer2))))
	c))

    ;; It is assumed that buf-length > 0.
    ;;
    ;; O: void
    (define (enlarge-buffer!)
      (if (> (* 2 buf-length) max-buf-length)
	  (error "Enlarge-buffer!" "maximum length reached, (length . max): "
		 (cons buf-length max-buf-length)))
      (let* ((new-buf-length  (* 2 buf-length))
	     (new-buf         (make-string new-buf-length))
	     (length-to-copy  (if (<= iend ibeg)
				  (- (string-length buf) (- iend ibeg))
				  (- iend ibeg))))
	;; buf is full: copy from buf to new-buf.
	(if (= 0 length-to-copy)
	    (error "enlager-buffer!" "buffer is empty. " (list iend ibeg (string-length buf))))
	
	(let loop ((i ibeg) (j 0))
	  (if (< j length-to-copy)
	      (if (< i buf-length)
		  (begin 
		    (string-set! new-buf j (string-ref buf i))
		    (loop (+ i 1) (+ j 1)))
		  (loop 0 j))))

	(set! buf new-buf)  ;; forget buf
	(set! buf-length new-buf-length)
	(set! ibeg 0)
	(set! iend length-to-copy)
	;; It was in state of read-in1 or read-in2; but now ibeg <
	;; iend so it is the state of read-in1.
	(set! cur-readc read-in1)))
    
    ;; All components are reversed.
    (define (snoc a b)  (cons a b))

  ;; To discard the characters in buffer without using any
  ;; operations. Used by the option 'discard' on a user rule.
  ;; I: void
  ;; O: void.
   (define (discard-chars)
    (if (not icur)
	;; Chars are not read from buffer. Therefore, all characters
        ;; that were stored can be discarded.
	(begin
	  (if (= ibeg iend)
	      (if no-char-used
		  (error "discard-chars"
			 "RegReg lexer terminated: recognized empty string twice."
			 "1")
		  (set! no-char-used #t))
	      (set! no-char-used #f))
	  ;; Just reset the buffer, forget its content.
	  (set! ibeg 0)
	  (set! iend 0))
	(begin
	  ;; Chars were read from the buffer. icur points the next
	  ;; char to read.
	  (if (= icur ibeg)
	      (if no-char-used
		  (error "discard-chars"
			 "RegReg lexer terminated: recognized empty string twice."
			 "2")
		  (set! no-char-used #t))
	      (set! no-char-used #f))
	    
	  (set! ibeg icur)
	  ;; Reset icur and cur-readc to appropriate values.
	  (if (= ibeg iend)
	      (begin ;; There are no characters to read from buffer.
		(set! icur #f)
		(set! cur-readc read-in1))
	      (begin ;; There are characters to read from buffer.
		;; icur = ibeg
		(cond
		 ((< ibeg iend)  (set! cur-readc read-from-buffer1))
		 (else (set! cur-readc read-from-buffer3))))))))

  ;; I: ops, [op].
  ;; O: 
  (define (apply-ops ops)
    
    ;; Index i points to the current character to read in buf.
    (let loop ((l ops) (i ibeg) (stack '()))
      (if (pair? l)
	  (let ((op (car l)))
	    (if (pair? op)
		(if (eqv? 'set (car op))
		    (loop (cdr l) i 
			  ;; bind an id with the top value on stack.
			  (cons (list (cadr op) (car stack))
				(cdr stack)))
		    (error "apply-ops" "unknown operation " op))
		(case op
		  ;; push_empty
		  ((2) (loop (cdr l) i (cons '() stack)))
		  ;; push
		  ((1)
		   (loop (cdr l) 
			 (if (= (+ i 1) buf-length) 0 (+ i 1)) 
			 (cons (string-ref buf i) stack)
			 ))
		  ;; snoc
		  ((0)
		   (let* ((a (car stack))
			  (b (cadr stack)))
		     (loop (cdr l) i (cons (snoc a b) (cddr stack)))))
		  ;; push snoc
		  ((3)
		   (let* ((b (car stack)))
		     (loop (cdr l) 
			   (if (= (+ i 1) buf-length) 0 (+ i 1)) 
			   (cons (snoc (string-ref buf i) b) (cdr stack)))))
		  (else 
		   (error "apply-ops" "unknown operation " op)))))
	  (begin ;; End of application, adjust the current read.
	    ;; 0 <= i < buf-length

	    ;; If no character is used from the buffer, record this
	    ;; fact for the next call to the lexer. Note: the buffer
	    ;; is never full. So if i=ibeg it is not due to wrap
	    ;; around of the above loop but simply because i has not
	    ;; increased from the code above.

	    (if (= i ibeg)
		(if no-char-used
		    (error "lexer"
			   "RegReg lexer terminated: recognized empty string twice."
			   "")
		    (set! no-char-used #t)))
	    
	    (set! ibeg i)
	    ;; Reset icur and cur-readc to appropriate values.
	    (if (= ibeg iend)
		(begin ;; There are no characters to read from buffer.
		  (set! icur #f)
		  (set! cur-readc read-in1))
		(begin ;; There are characters to read from buffer.
		  (set! icur i)
		  ;; icur = ibeg
		  (cond
		   ((< ibeg iend)  (set! cur-readc read-from-buffer1))
		   (else (set! cur-readc read-from-buffer3)))))
	    ;; Return result of the lexer
	    (car stack)))))

  ;; Return the buffer contents up to the point as a string.
  (define (buffer-contents)
    (if (< ibeg iend)
	(substring buf ibeg (or icur iend))
	(string-append (substring buf ibeg buf-length) 
		       (substring buf 0 (or icur iend)))))

  (define (reset-buffer!)
    (set! ibeg 0)
    (set! iend 0)
    (set! icur #f)
    (set! cur-readc read-in1))

  ;; I: v, vector as built by dfa->vector
  ;; O: (token result). token is a symbol; 
  ;;                    the result is a list or the user applied function to a list.
  (define (lexer v)
    (let* ((final-no-symb   #f) ;; (int symbol func type) the last final NFA node number 
                                ;; and token symbol recognized.
	   (last-final-icur #f)
	   (trace-trs       '())        ;; [(code-tr next-node-no [(no no [ops])])]
	   (final-tr-ptr    trace-trs)) ;; points in trace-trs for last final node

      ;; Return the list of rule tokens from the last taken transition.
      (define (get-rule-tokens)
	(if (pair? trace-trs)
	    (let* ((tr  (car trace-trs))
		   (no  (or (cadr tr) (and (pair? (cdr trace-trs))
					   (cadr (cadr trace-trs))))))
	      (if no
		  (va-tokens (vector-ref v no))
		  '()))
	    '()))

      ;; Return the new index on c. Register the transition in
      ;; trace-trs.
      ;;
      ;; I: c, char
      ;;    dn, va object
      ;; O: #f => there is no possible transition.
      ;;    int. The new index for vector v.
      (define (apply-trs c dn)
	
	;; I: c, character or eof-object.
	;;    tr, (transition . int)
	;;        transition is (u int (.. int int) (:: int ... int)
	;;                         (!:: int ... int) (.. int ... int))
	(define (tr-applicable? c tr)
	  (cond
	   ((eof-object? c)
	    (set! eof-reached #t)
	    (eqv? 'eof (car tr)))
	   ((integer? (car tr))
	    (= (char->integer c) (car tr)))
	   (else
	    (let ((a (char->integer c)))
	      (case (caar tr)
		((..)  (<= (cadar tr) a (caddar tr)))
		((!..) (not (<= (cadar tr) a (caddar tr))))
		((::)  (memv a (cdar tr)))
		((!::) (not (memv a (cdar tr))))
		((eof)    #f)
		(else (error "tr-applicable?" "Unknown transition " tr)))))))
	
	(let loop ((trs (va-trs dn)))
	  (if (pair? trs)
	      (if (tr-applicable? c (car trs))
		  (begin
		    (drvr:debug-trace "\nc: " c " tokens: " (va-tokens dn))
		    (trace-tr! (car trs))
		    (cadar trs))
		  (loop (cdr trs)))
	      #f
	      )))
      
      ;; I: (NFA-node-no symbol). The symbol correspond to a token id.
      ;; O: void
      (define (possible-final-node no-symb)
	(set! last-final-icur icur)
	(set! final-no-symb   no-symb)
	(set! final-tr-ptr    trace-trs))
      
      (define (trace-tr! tr)
	(set! trace-trs (cons tr trace-trs)))

      ;; The end of file has not been reached.
      (define (parse-input)
	;; Parse until no transition is applicable.
	(let loop ((c (cur-readc)) (cn (vector-ref v 1)))
	  (let* ((i   (apply-trs c cn))
		 (cnp (and i (vector-ref v i))))
	    (if cnp
		(begin
		  (if (va-final cnp)
		      (begin
			(possible-final-node (va-final cnp))))
		  (if (and (not eof-reached) (pair? (va-trs cnp)))
		      (loop (cur-readc) cnp))))))
	)

      ;; Begin lexer.

      ;; Trace the start node when entering the automaton.
      (trace-tr! (list #f 1 (va-start? (vector-ref v 1))))
      
      (if eof-reached
	  '(<eof> "")
	  (begin
	    ;; If the lexer recognized the empty string, the first
	    ;; node is final; this should be recorded.
	    (if (va-final (vector-ref v 1)) 
		(possible-final-node (va-final (vector-ref v 1))))
	    ;; Start main work of parsing.
	    (parse-input)
	    ;; A token was recognized or end of file was reached; or
	    ;; both or none.
	    (cond
	     (final-no-symb ;; a symbol has been recognized
	      (let* ((final-node-no (car   final-no-symb))
		     (token         (cadr  final-no-symb))
		     (func          (caddr final-no-symb))
		     (discard?      (list-ref final-no-symb 4)) 
		     (ops           (and (not discard?)
					 (build-ops-path dfa-vector final-node-no final-tr-ptr)))
		     (r        (or (and (not discard?)
					(tks-char->bindings (apply-ops ops))) 
				   '()))
		     (r2       (list token r))
		     (result   (if func ;; A func to apply to the result?
				   (func r2)
				   r2)))
		(if discard? (discard-chars))
		(drvr:debug-trace "\nLexer 1 result " result)
		result))
	     
	     (eof-reached ;; eof was not reached, but now it is.
	      (let* ((r      (buffer-contents))
		     (tokens (get-rule-tokens)))
		(if (pair? (cdr trace-trs)) 
		    (begin
		      ;; Eof was reached after reading at least one
		      ;; character without meeting a final state: This
		      ;; is considered an error.
		      (drvr:debug-trace "Lexer level 1; End of file; r:  " r 
					" tokens " tokens)
		      (reset-buffer!)
		      (if error-func
			  (error-func tokens r)
			  `(<error> ,tokens ,r)))
		    ;; Eof was reached but without reading any
		    ;; character.  This is not considered an error,
		    ;; even though there is no empty token recognized
		    ;; by the lexer; return eof indicator.
		    `(<eof> ,r))))
	     (else  ;; no basic token recognized.
	      (let* ((r      (buffer-contents))
		     (tokens (get-rule-tokens)))
		(drvr:debug-trace "\nNo basic token recognized.")
		(drvr:debug-trace "\nCurrent characters scanned are: `"r"'.")
		(drvr:debug-trace "\nCurrent active tokens: " tokens)
		(reset-buffer!)
		(if error-func
		    (error-func tokens r) 
		    `(<error> ,tokens ,r))))
	     )))))


  ;; -- Begin make-lexer --   
  (if (not (va-start? (vector-ref dfa-vector 1)))
      (error "lexer" "The first node is not a start node." dfa-vector))

  (init-buf)
  
  ;; Return a function to read the tokens. 
  (lambda () (lexer dfa-vector))
  ))

;; Construct the complete path of operations from a list
;; of DFA transitions.
;;
;; I: v, output of dfa->vector
;;    final-node, int. The NFA number of the final NFA node.
;;    trace-trs, [tr]. A tr is a transition: (code-tr next-node-no (no no [op]))
;;                     This list is in reverse order: from a final node to a start node.
;; O: [op]
;;
(define (build-ops-path v final-node trace-trs)
  (let loop ((l trace-trs) (last-no final-node) (ops '()))
    (if (pair? l)
	(let* ((tr       (car l))
	       ;; Every p is (no-begin no-end ops-path)
	       (new-tr   (p-exists? (lambda (p) (= last-no (cadr p)))
				    (caddr tr))))
	  (if new-tr
	      ;; Add the operations of this arc.
	      (loop (cdr l) (car new-tr) (append (caddr new-tr) ops))
	      (error "build-ops-path" "internal error, no new-tr!!!" (list last-no trace-trs v))))
	;; Do not reverse ops, as the operations are now in the right order.
	ops)))

;; I: (u str 
;;       (sym [(sym [binding])]))
;;     a binding is a list of char or (sym char ... char)
;; O: (sym [(sym [(sym str) or str])])
(define (tks-char->bindings tks-chars)

  ;; I: s, [char]
  (define (chars->strings cs r s)
    (cond 
     ((pair? cs)
      (cond 
       ((symbol? (car cs))  (cons (car cs) (chars->strings (cdr cs) '() '())))
       ((char? (car cs))    (chars->strings (cdr cs) r (cons (car cs) s)))
       ((or (pair? (car cs)) (null? (car cs)))
	(if (null? s)
	    (chars->strings (cdr cs) 
			    (cons (chars->strings (car cs) '() '()) r)
			    '())
	    (chars->strings (cdr cs) 
			    (cons (chars->strings (car cs) '() '()) 
				  (cons (list->string s)
					r))
			    '())))
       (else (error "chars->strings" "unkown case " (car cs)))))
     ((char? cs)
      (string cs))
     (else
      (if (null? s)
	  r
	  (cons (list->string s) r)))))
  (chars->strings tks-chars '() '()))


(define (drvr:list-string-append l)
  (if (pair? l)
      (string-append (car l) (drvr:list-string-append (cdr l)))
      ""
      ))

;; Similar to list-string-append, but add a space between each string.
(define (drvr:strings->string l)
  (cond 
   ((and (pair? l) (pair? (cdr l)))
    (string-append (car l) " " (drvr:strings->string (cdr l))))
   ((pair? l) (car l))
   (else "")))

;; (foldr f a (x1 x2 ... xn)) = (f x1 (f x2 ... (f xn a) ... ))
;;

(define (foldli f a l i)
  (if (pair? l)
      (foldli f (f i a (car l)) (cdr l) (+ i 1))
      a))

(define (p-first-one-of p l)
  (if (pair? l)
      (or (p (car l))(p-first-one-of p (cdr l)))
      #f))

(define (p-exists? p l)
  (if (pair? l)
      (or (if (p (car l)) (car l) #f) (p-exists? p (cdr l)))
      #f))

(define (display-l . l) (for-each display l))
(define (display-ln . l) (for-each display l) (newline))

;; These functions must be as in regreg.scm.
(define (va-trs v)               (vector-ref v 0))
(define (va-start? v)            (vector-ref v 1))
(define (va-final  v)            (vector-ref v 2))
(define (va-tokens v)            (vector-ref v 3))
(define (node-final-discard? n) 
  (and (pair? (node-final n)) (eqv? 'discard (list-ref (node-final n) 4))))
(define (node-final     n)       (vector-ref  n 2))
